import re
import traceback
from .classification_exceptions import SelectionOutOfRange, ReversedRange, UnbalancedPageRange, NotZeroIndex


"""
Functionality for parsing the page ranges for 'MULTIPLE' document classification scenarios.

Who knows what the annotators will enter, so we account for all of the basic scenarios via RegEx.
"""


def parse_page_range(page_range: str, max_pages: int):
    """
    Parses the page range coming from the annotator for a multi-classification segment.
    """
    func = parse_page_range.__name__
    try:
        page_range = re.sub(" ", "", page_range)
        print(f"[{func}] input={page_range}")
        if bool(re.compile(r"(\d+|\d)-(\d+|\d)((,(\d+|\d))+)").search(page_range)):
            """
            Example: input = '1-5, 8, 9, 11'
            """
            res = re.compile(r"(\d+|\d)-(\d+|\d)((,(\d+|\d))+)").search(page_range)
            if res is not None:
                print(f"[{func}] option 1 matched")
                start = int(res.group(1))
                end = int(res.group(2))
                others_ = res.group(3)
                others = others_.split(",")
                others = [int(i) for i in others if bool(re.compile(r"\d+|\d").search(i))]
                print(f"start={start}; end={end}; others={others}")

                nums = others.copy()
                nums.extend([start, end])
                highest = max(nums)
                if highest > max_pages:
                    raise SelectionOutOfRange(
                        f"The page range selection exceeds the maximum number of pages; {highest} > {max_pages}")

                if start > end:
                    raise ReversedRange(f"Out of order; {start} is bigger than {end}")

                if start < max(others) <= end:
                    raise UnbalancedPageRange(f"The largest comma-separated-value {max(others)} is included in the range {start}-{end}")

                for num in others:
                    if start < num <= end:
                        raise UnbalancedPageRange(f"Comma-separated-value {num} is included in the range {start}-{end}")

                if start == 0 or 0 in others:
                    raise NotZeroIndex("Page ranges and indexes do not include '0'")

                return True, {
                    'range_start': start,
                    'range_end': end,
                    'singles': sorted(others)
                }, ""

        elif bool(re.compile(r"(((\d+|\d),)+)(\d+|\d)-(\d+|\d)").search(page_range)):
            """
            Example: input = '1, 2, 4-7'
            """
            res = re.compile(r"(((\d+|\d),)+)(\d+|\d)-(\d+|\d)").search(page_range)
            if res is not None:
                print(f"[{func}] option 2 matched")
                start = int(res.group(4))
                end = int(res.group(5))
                others_ = res.group(1)
                others = others_.split(",")
                others = [int(i) for i in others if bool(re.compile(r"\d+|\d").search(i))]
                print(f"start={start}; end={end}; others={others}")

                nums = others.copy()
                nums.extend([start, end])
                highest = max(nums)
                if highest > max_pages:
                    raise SelectionOutOfRange(
                        f"The page range selection exceeds the maximum number of pages; {highest} > {max_pages}")

                if start > end:
                    raise ReversedRange(f"Out of order; {start} is bigger than {end}")

                if start < max(others) <= end:
                    raise UnbalancedPageRange(f"The largest comma-separated-value {max(others)} is included in the range {start}-{end}")

                for num in others:
                    if start < num <= end:
                        raise UnbalancedPageRange(f"Comma-separated-value {num} is included in the range {start}-{end}")

                if start == 0 or 0 in others:
                    raise NotZeroIndex("Page ranges and indexes do not include '0'")

                return True, {
                    'range_start': start,
                    'range_end': end,
                    'singles': sorted(others)
                }, ""

        elif bool(re.compile(r"(\d+|\d)-(\d+|\d)").search(page_range)):
            """
            Example: input = '1-3'
            """
            res = re.compile(r"(\d+|\d)-(\d+|\d)").search(page_range)
            if res is not None:
                print(f"[{func}] option 3 matched")
                start = int(res.group(1))
                end = int(res.group(2))
                print(f"start={start}; end={end}")

                if start > end:
                    raise ReversedRange(f"Out of order; {start} is bigger than {end}")

                if end > max_pages:
                    raise SelectionOutOfRange(
                        f"The page range selection exceeds the maximum number of pages; {end} > {max_pages}")

                if start == 0:
                    raise NotZeroIndex("Page ranges and indexes do not include '0'")

                if start == end:
                    print(f"{start} == {end}")
                    return 200, {
                        'range_start': None,
                        'range_end': None,
                        'singles': [start],
                        'error': None
                    }

                return True, {
                    'range_start': start,
                    'range_end': end,
                    'singles': []
                }, ""

        elif bool(re.compile(r"(((\d+|\d),?)+)").search(page_range)):
            """
            Example: input = '1, 2, 3'
            """
            res = re.compile(r"(((\d+|\d),?)+)").search(page_range)
            if res is not None:
                print(f"[{func}] option 4 matched")
                others_ = res.group(1)
                others = others_.split(",")
                others = [int(i) for i in others if bool(re.compile(r"\d+|\d").search(i))]
                print(f"others={others}")

                highest = max(others)
                if highest > max_pages:
                    raise SelectionOutOfRange(
                        f"The page range selection exceeds the maximum number of pages; {highest} > {max_pages}")

                if 0 in others:
                    raise NotZeroIndex("Page ranges and indexes do not include '0'")

                return True, {
                    'range_start': None,
                    'range_end': None,
                    'singles': sorted(others)
                }, ""

        elif bool(re.compile(r"(\d+|\d)").search(page_range)):
            res = re.compile(r"(\d+|\d)").search(page_range)
            if res is not None:
                print(f"[{func}] option 5 matched")
                start = int(res.group(1))
                print(f"single start={start}")

                if start > max_pages:
                    raise SelectionOutOfRange(
                        f"The page range selection exceeds the maximum number of pages; {start} > {max_pages}")

                if start == 0:
                    raise NotZeroIndex("Page ranges and indexes do not include '0'")

                return True, {
                    'range_start': None,
                    'range_end': None,
                    'singles': [start]
                }, ""

    except UnbalancedPageRange as ub:
        print(f"[INPUT ERROR] UnbalancedPageRange ==> {ub}")
        return False, {
            'range_start': None,
            'range_end': None,
            'singles': []
        }, f"UnbalancedPageRange ==> {ub}"
    except SelectionOutOfRange as sor:
        print(f"[INPUT ERROR] SelectionOutOfRange ==> {sor}")
        return False, {
            'range_start': None,
            'range_end': None,
            'singles': []
        }, f"SelectionOutOfRange ==> {sor}"
    except ReversedRange as rr:
        print(f"[INPUT ERROR] ReversedRange ==> {rr}")
        return False, {
            'range_start': None,
            'range_end': None,
            'singles': []
        }, f"ReversedRange ==> {rr}"
    except NotZeroIndex as nzi:
        print(f"[INPUT ERROR] NotZeroIndex ==> {nzi}")
        return False, {
            'range_start': None,
            'range_end': None,
            'singles': []
        }, f"NotZeroIndex ==> {nzi}"
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, {
            'range_start': None,
            'range_end': None,
            'singles': []
        }, f"{e}"



"""

schema kung-fu

"""
from . import accessorial, bill_of_lading, classification, classification_exceptions, invoice, noa_lor, \
    parse_classification_page_range, rate_confirmation, rsi_translator, schema_utilities



import re
import traceback
import json
import decimal
from .parse_classification_page_range import parse_page_range


def get_main_schema():
    main_schema = {
        "headers": {
            "Classify Document": {
                "classification": "N/A"
            }
        },
        "tables": {
            "Document Classifications": [
                {
                    "Page Rows": [
                        {"doc_type": "N/A", "page_nums": "0"},
                        {"doc_type": "N/A", "page_nums": "0"}
                    ]
                }
            ]
        }
    }
    return main_schema


DT_CLASSIFICATION_MAP = {
    "rate_confirmation": "Rate Confirmation (ratecon)",
    "bill_of_lading": "Bill of Lading (BOL)",
    "noa_lor": "Notice of Assignment (NOA)",
    "noa": "Notice of Assignment (NOA)",
    "lor": "Letter of Rescission (LOR)",
    "scale_ticket": "Scale Ticket",
    "packing_list": "Packing List",
    "drayage": "Drayage",
    "chassis": "Chassis",
    "detention": "Detention",
    "lumper": "Lumper",
    "other": "OTHER",
    "receipt": "Receipt",
    "invoice": "Invoice",
    None: "N/A"
}


def dt_multiclass_to_dynamoson(autoextracts_data, page_length):
    """
    Reformats the DTs multiclass data format into the mutually agreed upon Dynamoson structure;
    :param autoextracts_data: dict,;
    :param page_length: int,;
    :return:
    """
    func = dt_multiclass_to_dynamoson.__name__
    base_schema = {
        "headers": {
            "Classify Document": {
                "classification": "MULTIPLE"
            }
        },
        "tables": {
            "Document Classifications": [
                {
                    "Page Rows": []
                }
            ]
        }
    }
    try:
        for num, class_dict in autoextracts_data.items():
            pages = [str(i) for i in class_dict['singles']]
            page_nums = ", ".join(pages)
            doc_class = class_dict['classification']
            try:
                doc_class = DT_CLASSIFICATION_MAP[doc_class]
            except Exception as ee:
                print(f"[WARN] {doc_class} not found in DT_CLASSIFICATION_MAP ==> {ee}")

            row = {
                'doc_type': doc_class,
                'page_nums': page_nums
            }
            base_schema['tables']['Document Classifications'][0]['Page Rows'].append(row)
        try:
            sorted_results = sorted(base_schema['tables']['Document Classifications'][0]['Page Rows'], key=lambda k: max([int(i) for i in k['page_nums']]))
            print(f"[DEBUG] sorted preclassifications: {sorted_results}")
            base_schema['tables']['Document Classifications'][0]['Page Rows'] = sorted_results
        except Exception as se:
            print(f"[WARN] Sort Error ==> {se}")
            traceback.print_exc()

        if len(base_schema['tables']['Document Classifications'][0]['Page Rows']) == 0:
            for i in list(range(0, page_length)):
                base_schema['tables']['Document Classifications'][0]['Page Rows'].append(
                    {
                        "doc_class": "N/A",
                        "page_nums": str(i + 1)
                    }
                )

        return True, base_schema, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, base_schema, f"{e}"


def classification_template(doc, multi_class=False):
    func = classification_template.__name__
    try:
        assert isinstance(doc, dict), f"Expected dictionary DataType for document, but got <{type(doc)}> instead."

        if multi_class:
            print(f"[PARSING PRE-CLASSIFIED PAGES] {doc['autoextracts']['data']}")
            success, dynamoson, err = dt_multiclass_to_dynamoson(
                autoextracts_data=doc["autoextracts"]["data"],
                page_length=len(doc['image_filenames'])
            )
            if not success:
                raise Exception(f"dt_multiclass_to_dynamoson() encountered an error ==> {err}")
            ddb_copy = doc.copy()
            ddb_copy["autoextracts"]["data"] = dynamoson
            return True, ddb_copy, ""

        else:
            ddb_copy = doc.copy()
            ddb_copy["autoextracts"]["data"] = get_main_schema()
            return True, ddb_copy, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"


def standardize_classification_name(classification: str):
    """
    Takes in a string classification choice from the DT and parses it into a standard format.
    """
    func = standardize_classification_name.__name__
    try:
        class_names = [
            "rate_confirmation", "bill_of_lading", "noa_lor", "packing_list", "scale_ticket", "invoice", "receipt",
            "chassis", "detention", "drayage", "other", "multiple", "lumper"
        ]

        if classification.lower() in ["ratecon", "rate con", "rate confirmation", "rate_confirmation", "rate confirmation (ratecon)"]:
            classification = "rate_confirmation"
        elif "ratecon" in classification.lower() or "rate con" in classification.lower():
            classification = "rate_confirmation"
        elif classification.lower() in ["bol", "bill of lading", "bill_of_lading", "bill of lading (bol)"]:
            classification = "bill_of_lading"
        elif "bill of lading" in classification.lower() or "bol" in classification.lower():
            classification = "bill_of_lading"
        elif classification.lower() in ["noa", "lor", "noa-lor", "noalor", "noa_lor"]:
            classification = "noa_lor"
        elif "noa" in classification.lower() or "lor" in classification.lower():
            classification = "noa_lor"
        elif classification.lower() in ["packing list", "packing-list", "packing_list"]:
            classification = "packing_list"
        elif classification.lower() == "invoice":
            classification = classification.lower()
        elif classification.lower() == "lumper":
            classification = classification.lower()
        elif classification.lower() == "receipt":
            classification = classification.lower()
        elif classification.lower() in ["scale ticket", "scale_ticket", "scale-ticket"]:
            classification = "scale_ticket"
        elif classification.lower() == "drayage":
            classification = classification.lower()
        elif classification.lower() in ["chassis", "chaissis"]:
            classification = "chassis"
        elif classification.lower() == "detention":
            classification = classification.lower()
        elif classification.lower() in ["other", "na", "n/a"]:
            classification = "other"
        elif classification.lower() == "multiple":
            classification = classification.lower()

        if classification not in class_names:
            raise Exception(f"[WARN] {classification} not accounted for!")

        return True, classification, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"


def parse_classification_dynamoson(flattened_ddb_entry: str) -> tuple[bool, str, str]:
    func = parse_classification_dynamoson.__name__
    try:
        if isinstance(flattened_ddb_entry, dict):
            flattened_ddb_entry = json.dumps(flattened_ddb_entry)
        data = json.loads(flattened_ddb_entry, parse_float=decimal.Decimal)
        print(f"[{func}] dynamoson: {data}")
        classification_ = data['headers']['Classify Document']['classification']
        stand_success, classification, stand_err = standardize_classification_name(classification_)
        if not stand_success:
            raise Exception(f"standardize_classification_name() >>> {stand_err}")
        print(f"[{func}] returning {classification}")
        return True, classification, ""
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, "None", f"{e}"


def parse_multiple_classification_dynamoson(dynamoson: str, ddb_doc_entry, req_id) -> tuple[bool, dict, str]:
    """
    When the ['headers']['Classify Document']['classification'] of the Dynamoson == "MULTIPLE",
    Re-parse the dynamoson to look at the ['tables']['Document Classifications'] (array of objects)
    to get the doc-by-doc classification for this concatenated image.
    """
    func = parse_multiple_classification_dynamoson.__name__
    classification_dict = {}
    try:
        if isinstance(dynamoson, dict):
            dynamoson = json.dumps(dynamoson)

        if isinstance(ddb_doc_entry, str):
            ddb_doc_entry = json.loads(ddb_doc_entry)

        image_filenames = ddb_doc_entry['image_filenames']
        total_pages = len(image_filenames)
        data = json.loads(dynamoson, parse_float=decimal.Decimal)
        print(f"[{func}] dynamoson: {data}")
        classifications = data['tables']['Document Classifications'][0]['Page Rows']  # list of dicts
        print(f"[DEBUG] classifications: {classifications}")

        if total_pages > len(classifications):
            raise Exception(f"Each page requires it's own classification row! Number of pages: {total_pages}; Submitted Classifications: {len(classifications)}")
        elif total_pages < len(classifications):
            raise Exception(f"More classifications detected than there are pages! Classifications: {len(classifications)}; Pages: {total_pages}")

        ordinal = 1
        for class_dict in classifications:
            print(f"[parsing multiple classification {ordinal}] {class_dict}")

            classification_ = class_dict['doc_type']
            page_range_string = class_dict['page_nums']

            if classification_ is None or (isinstance(classification_, str) and classification_.lower() == "n/a"):
                raise Exception(f"Classification of {classification_} not valid. Please classify the page using the drop-down menu options.")

            stand_success, classification, stand_err = standardize_classification_name(classification_)
            if not stand_success:
                raise Exception(f"standardize_classification_name() >>> {stand_err}")

            success, page_range_dict, err = parse_page_range(page_range=page_range_string, max_pages=total_pages)
            if not success:
                raise Exception(f"parse_page_range() >>> {err}")

            page_res = page_range_dict.copy()
            page_res.update({'classification': classification})

            classification_dict.update({
                str(ordinal): page_res
            })

            ordinal += 1

        # Double-check the data by translating it
        translation_success, translated_data, translation_err = translate_dt_multiclass(
            req_id=req_id,
            data=classification_dict,
            total_pages=total_pages
        )
        if not translation_success:
            anno_errors = [
                "not chronologically related",
                "used more than once",
                "does not match up"
            ]
            for a_err in anno_errors:
                if a_err in translation_err:
                    raise Exception(translation_err)

            raise Exception(f"translate_dt_multiclass() >>> {translation_err}")

        return True, classification_dict, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        classification_dict.update({
            'error': f"{e}"
        })
        return False, classification_dict, f"{e}"


def translate_dt_multiclass(req_id, data, total_pages) -> tuple[bool, dict, str]:
    """
    Translates the DTs multiclass output into something similar to what the auto-classifier would do if it
    succeeded all the way through; This format has some benefits, such as easily determining if all document pages
    were properly accounted for, as well as translating into other areas such as splitting into separate documents;
    :param req_id:
    :param data:
    :param total_pages:
    :return:
    """
    func = translate_dt_multiclass.__name__
    """
    DT multiclass format:::
    {
        "1": {
                'classification': {doc_class},
                'range_start': int|None,
                'range_end': int|None,
                'singles': list(<int>|empty)
        },
        "2": {
                'classification': {doc_class},
                'range_start': int|None,
                'range_end': int|None,
                'singles': list(<int>|empty)
        },
        etc...
    }
    """
    try:
        page_numbers = list()
        translated = {}
        for num, d, in data.items():

            nums = list()
            if d.get('range_start') is not None and d.get('range_end') is not None:
                nums = [i for i in list(range(int(d['range_start']), int(d['range_end']) + 1))]
            if len(d['singles']) >= 1:
                nums.extend([int(i) for i in d['singles']])

            page_numbers.extend(nums)

            for n in nums:
                if str(n) in translated.keys():
                    raise Exception(f"Page number <{n}> appears to be used more than once! {translated}")

                translated.update({
                    str(n): {
                        "doc_class": d['classification'],
                        "model_class": None
                    }
                })
        page_numbers = sorted(page_numbers)
        first_page = page_numbers[0]
        addition = 1
        for p in page_numbers:
            if page_numbers.index(p) == 0:
                continue
            elif first_page + addition != p:
                raise Exception(f"Flattened page numbers are not chronologically related! page_numbers = {page_numbers}")
            else:
                addition += 1

        if len(page_numbers) != total_pages:
            raise Exception(f"Total number of pages does not match up! Total Pages: {total_pages}; Parsed Annotation Pages: {page_numbers}")

        return True, translated, ""

    except Exception as e:
        print(f"[({req_id}) ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, {}, f"{e}"



import datetime
import json
import re
import traceback
from .schema_utilities import get_idtype, parse_incoming_address, parse_time, format_state


def get_date_schema():
    return {
        "date": None,
        "time": None,
        "datetype": None
    }


def get_reference_schema(idtype):
    _idtype = get_idtype(idtype)
    return {
        "id": None,
        "idtype": idtype,
        "_idtype": _idtype
    }


def get_lineitem_schema():
    return {
        "amount": "0.0",  # Total amount (rate * qty)
        "rate": "0.0",  # Price each
        "qty": "0",
        "id": None  # Name / description
    }


def get_entity_schema(type_):
    func = get_entity_schema.__name__
    try:
        type_code = None
        if type_ == "SHIPPER":
            type_code = "SH"
        elif type_ == "CARRIER":
            type_code = "CR"
        elif type_ == "CONSIGNEE":
            type_code = "CN"
        elif type_ == "BROKER":
            type_code = "BK"
        elif type_ == "BILL TO":
            type_code = "BT"
        elif type_ == "BILL FROM":
            type_code = "BF"

        if type_code is None:
            raise Exception(f"Entity Type <{type_}> not provisioned.")

        return {
            "name": None,
            "type": type_,
            "_type": type_code,
            "id": None,
            "_idtype": "ZZ",
            "address": [],
            "city": None,
            "state": None,
            "postal": None,
            "contacts": {
                "contactname": None,
                "contact_type": None,
                "contact_number": None,
                "contact_number_type": None
            }
        }

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        return None


def get_base_invoice_schema():
    return {
        "identifier": None,
        "invoice": {
            "terms": None,
            "total": None,
            "line_items": []
        },
        "dates": [],
        "references": [],
        "entities": []
    }


def get_invoice_fail_schema():
    base = get_base_invoice_schema()
    base['dates'] = [get_date_schema()]
    base['references'] = [get_reference_schema("REFERENCE NUMBER")]
    base['entities'] = [
        get_entity_schema("SHIPPER"),
        get_entity_schema("CONSIGNEE"),
        get_entity_schema("BILL FROM"),
        get_entity_schema("BILL TO")
    ]
    return base


def get_flat_line_item_schema():
    return {
        "amount": None,
        "rate": None,
        "quantity": None,
        "name": None
    }


def get_flat_date_schema():
    return {
        "date_0": None,
        "datetype_0": None,
        "time_0": None
    }


def get_flat_reference_schema(idtype=None):
    return {
        "id": None,
        "idtype": idtype
    }


def get_flat_entity_schema(type_=None):
    return {
        "type": type_,
        "name": None,
        "address": None,
        "city": None,
        "state": None,
        "postal": None,
        "country": None,
        "contactname": None,
        "contact_type": None,
        "contact_number": None,
        "contact_number_type": None
    }


def get_flat_base_schema():
    """
    Flat Invoice Schema.
    """
    return {
        "headers": {
            "General": {
                "identifier": None,
                "terms": None,
                "total": None
            }
        },
        "tables": {
            "General Info": [
                {
                    "Line Items": [],
                    "Document Dates": [],
                    "Document References": [],
                    "Document Entities": []
                }
            ]
        }
    }


def blank_schema():
    base = get_flat_base_schema()
    base['tables']['General Info'][0]['Line Items'] = [get_flat_line_item_schema()]
    base['tables']['General Info'][0]['Document Dates'] = [get_flat_date_schema()]
    base['tables']['General Info'][0]['Document References'] = [get_flat_reference_schema("REFERENCE NUMBER")]
    base['tables']['General Info'][0]['Document Entities'] = [
        get_flat_entity_schema("SHIPPER"),
        get_flat_entity_schema("CONSIGNEE"),
        get_flat_entity_schema("BILL FROM"),
        get_flat_entity_schema("BILL TO")
    ]
    return base


def flatten_invoice(ddb_entry) -> tuple[bool, dict, str]:
    """
    Translate Invoice KVT into Dynamoson.
    """
    func = flatten_invoice.__name__
    try:
        if isinstance(ddb_entry, str):
            ddb_entry = json.loads(ddb_entry)
        ddb_entry = ddb_entry.copy()

        send_blank = False
        auto_ex = ddb_entry.get('autoextracts')

        if auto_ex is None:
            print(f"[WARN] 'ddb_entry' has no 'autoextracts' object")
            send_blank = True

        elif auto_ex.get('data') is None:
            print(f"[WARN] 'ddb_entry'[autoextracts] has no 'data' object")
            send_blank = True

        elif len(auto_ex['data'].keys()) == 0:
            print(f"[WARN] 'ddb_entry'[autoextracts][data] has a key-length of 0")
            send_blank = True

        if send_blank:
            blank = blank_schema()
            ddb_entry.update({
                "autoextracts": {
                    "data": blank
                }
            })
            return True, ddb_entry, ""

        ### Begin mapping data from the KVT to the Dynamoson
        data = auto_ex["data"]
        main = get_flat_base_schema()

        main['headers']['General']['identifier'] = data['identifier']
        main['headers']['General']['terms'] = data['invoice']['terms']
        main['headers']['General']['total'] = data['invoice']['total']

        line_items = list()
        for line in data['invoice']['line_items']:
            li = get_flat_line_item_schema()
            li.update({
                "name": line['id'],
                "amount": str(line['amount']),
                "rate": str(line['rate']),
                "quantity": str(line['qty'])
            })
            line_items.append(li)

        if len(line_items) == 0:
            line_items.append(get_flat_line_item_schema())

        main['tables']['General Info'][0]['Line Items'] = line_items

        dates = list()
        for date in data['dates']:
            d = get_flat_date_schema()
            d.update({
                "date_0": date['date'],
                "datetype_0": date['datetype'],
                "time_0": date['time']
            })
            dates.append(d)

        if len(dates) == 0:
            dates.append(get_flat_date_schema())

        main['tables']['General Info'][0]['Document Dates'] = dates

        entities = list()
        for entity in data['entities']:
            ent = get_flat_entity_schema()
            ent.update({
                "type": entity['type'],
                "name": entity['name'],
                "address": entity['address'],
                "city": entity['city'],
                "state": entity['state'],
                "postal": entity['postal'],
                "contactname": entity['contacts']['contactname'],
                "contact_type": entity['contacts']['contact_type'],
                "contact_number": entity['contacts']['contact_number'],
                "contact_number_type": entity['contacts']['contact_number_type']
            })
            entities.append(ent)

        if len(entities) == 0:
            entities.append(get_flat_entity_schema())

        main['tables']['General Info'][0]['Document Entities'] = entities

        references = list()
        for ref in data['references']:
            r = get_flat_reference_schema()
            r.update({
                "id": ref['id'],
                "idtype": ref['idtype']
            })
            references.append(r)

        if len(references) == 0:
            references.append(get_flat_reference_schema())

        main['tables']['General Info'][0]['Document References'] = references

        flattened_data = json.loads(json.dumps(main))
        ddb_entry['autoextracts']['data'] = flattened_data
        return True, ddb_entry, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, {}, f"{e}"


def unflatten_invoice(dynamoson) -> tuple[bool, dict, str]:
    """
    Translate flattened Invoice data back into original KVT schema.
    """
    func = unflatten_invoice.__name__
    if isinstance(dynamoson, str):
        dynamoson = json.loads(dynamoson)

    try:
        main = get_base_invoice_schema()
        main["identifier"] = dynamoson['headers']['General']['identifier']
        total = dynamoson['headers']['General']['total']
        total = re.sub(r"[^0-9.]+", "", total)
        main['invoice'].update({
            "terms": dynamoson['headers']['General']['terms'],
            "total": total
        })

        line_items = list()
        for line in dynamoson['tables']['General Info'][0]['Line Items']:
            amount = line['amount']
            amount = re.sub(r"[^0-9.]+", "", amount)
            rate = line['rate']
            rate = re.sub(r"[^0-9.]+", "", rate)
            qty = line['quantity']
            qty = re.sub(r"[^0-9.]+", "", qty)
            l = get_lineitem_schema()
            l.update({
                "id": line['name'],
                "amount": amount,
                "rate": rate,
                "qty": qty
            })
            line_items.append(l)

        main['invoice']['line_items'] = line_items

        dates = list()
        for date in dynamoson['tables']['General Info'][0]['Document Dates']:
            d = get_date_schema()

            date_ = date['date_0']
            if date_ is not None:
                date_success, date_, date_err = parse_date(date_str=date_)
                if not date_success:
                    raise Exception(f"Date in Document Dates index {dynamoson['tables']['General Info'][0]['Document Dates'].index(date)} is incorrectly formatted; date={date['date_0']}")

            time_ = date['time_0']
            if time_ is not None:
                time_success, time_, time_err = parse_time(time_str=time_)
                if not time_success:
                    raise Exception(f"Time in Document Dates index {dynamoson['tables']['General Info'][0]['Document Dates'].index(date)} is incorrectly formatted; time={date['time_0']}")

            d.update({
                "date": date_,
                "datetype": date['datetype_0'],
                "time": time_
            })
            dates.append(d)
        main['dates'] = dates

        references = list()
        for ref in dynamoson['tables']['General Info'][0]['Document References']:
            r = get_reference_schema(idtype=ref['idtype'])
            r.update({
                "id": ref['id']
            })
            references.append(r)
        main['references'] = references

        entities = list()
        for ent in dynamoson['tables']['General Info'][0]['Document Entities']:
            entity = get_entity_schema(type_=ent['type'])

            name = ent['name']
            if name is not None:
                name = name.title()

            city = ent['city']
            if city is not None:
                city = city.title()

            state = ent['state']
            if state is not None:
                state_success, state, state_err = format_state(
                    state=state,
                    abbreviate=True
                )
                if not state_success:
                    raise Exception(f"{state_err}")

            address = []
            if ent['address'] is not None:
                address_success, address, address_err = parse_incoming_address(
                    address=ent['address']
                )
                if not address_success:
                    raise Exception(f"parse_incoming_address() >>> {address_err}")
                if bool(re.compile(r"\S+").search(str(address))):
                    address = [address]
                else:
                    address = []

            uid_values = [i for i in [name, city, state, ent['postal']] if i is not None]
            uid_ = "".join(uid_values)
            uid = re.sub('[^A-Za-z0-9]+', '', uid_)
            entity.update({
                "name": name,
                "id": uid,
                "city": city,
                "state": state,
                "postal": ent['postal'],
                "address": address,
                "contacts": {
                    "contactname": ent['contactname'],
                    "contact_type": ent['contact_type'],
                    "contact_number": ent['contact_number'],
                    "contact_number_type": ent['contact_number_type']
                }
            })
            entities.append(entity)
        main['entities'] = entities

        return True, main, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, {}, f"{e}"


def parse_date(date_str):
    func = parse_date.__name__
    try:
        date_ = datetime.datetime.strptime(date_str, '%Y-%m-%d')
        date_ = date_.strftime('%Y-%m-%d')
        return True, date_, ""
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"





import re
import json
import traceback
from .schema_utilities import get_idtype, format_time


"""
The following schemas and translation functions are intended to be used for any document that falls into the 
'Accessorial' category >> Lumper, Drayage, Chassis, & Detention
"""


def get_flat_accessorial_schema():
    schema = {
        "headers": {
            "General": {
                "identifier": "",
                "amount": ""
            }
        },
        "tables": {
            "General Info": [
                {
                    "Document Dates": [],
                    "Document References": []
                }
            ]
        }
    }
    return schema


def get_flat_date_schema():
    return {
        "date_0": "",
        "datetype_0": "",
        "time_0": ""
    }


def get_flat_reference_schema():
    return {
        "id": "",
        "idtype": ""
    }


def blank_schema():
    blank = get_flat_accessorial_schema()
    blank['tables']['General Info'][0].update({
        "Document Dates": [get_flat_date_schema()],
        "Document References": [get_flat_reference_schema()]
    })
    return blank


def flatten_accessorial(ddb_entry) -> tuple[bool, dict, str]:
    """
    Translates an Accessorial document (Lumper, Chassis, Detention, Drayage) KVT into Dynamoson.
    """
    func = flatten_accessorial.__name__
    if isinstance(ddb_entry, str):
        ddb_entry = json.loads(ddb_entry)
    ddb_entry = ddb_entry.copy()
    try:
        auto_ex = ddb_entry.get('autoextracts')

        send_blank = False
        if auto_ex is None:
            print(f"[WARN] 'ddb_entry' has no 'autoextracts' object")
            send_blank = True
        elif auto_ex.get('data') is None:
            print(f"[WARN] 'ddb_entry'[autoextracts] has no 'data' object")
            send_blank = True
        elif len(auto_ex['data'].keys()) == 0:
            print(f"[WARN] 'ddb_entry'[autoextracts][data] has a key-length of 0")
            send_blank = True
        if send_blank:
            ddb_entry.update({
                "autoextracts": {
                    "data": blank_schema()
                }
            })
            return True, ddb_entry, ""

        data = auto_ex["data"]
        main = get_flat_accessorial_schema()

        identifier = str(data['identifier'])
        amount = str(data['amount'])
        amount = re.sub(r"[^\d,.]+", "", amount)
        main["headers"]["General"].update({
            "identifier": identifier,
            "amount": amount
        })

        dates = data.get('dates')
        if len(dates) == 0:
            main["tables"]["General Info"][0].update({
                "Document Dates": [get_flat_date_schema()]
            })
        else:
            for d in dates:
                date_dict = get_flat_date_schema()
                date_dict.update({
                    "date_0": d['date'],
                    "datetype_0": "lumper",
                    "time_0": d['time']
                })
                main["tables"]["General Info"][0]["Document Dates"].append(date_dict)

        references = data.get('references')
        if len(references) == 0:
            main["tables"]["General Info"][0].update({
                "Document References": [get_flat_reference_schema()]
            })
        else:
            for r in references:
                ref_dict = get_flat_reference_schema()
                ref_dict.update({
                    "id": r['id'],
                    "idtype": r['idtype']
                })
                main["tables"]["General Info"][0]["Document References"].append(ref_dict)

        ddb_entry.update({
            "autoextracts": {
                "data": main
            }
        })
        return True, ddb_entry, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        ddb_entry = ddb_entry.copy()
        ddb_entry.update({
            "autoextracts": {
                "data": blank_schema()
            }
        })
        return False, ddb_entry, f"{e}"


def get_date_schema():
    return {
        "date": None,
        "datetype": None,
        "time": None
    }


def get_reference_schema(idtype):
    _idtype = get_idtype(idtype)
    return {
        "id": None,
        "idtype": idtype,
        "_idtype": _idtype
    }


def get_accessorial_schema():
    return {
        "identifier": None,
        "amount": None,
        "dates": [],
        "references": []
    }


def unflatten_accessorial(dynamoson) -> tuple[bool, dict, str]:
    """
    Translates flattened Accessorial document (Lumper, Chassis, Detention, Drayage) back into KVT schema.
    """
    func = unflatten_accessorial.__name__
    if isinstance(dynamoson, str):
        dynamoson = json.loads(dynamoson)
    try:
        main = get_accessorial_schema()
        identifier = dynamoson["headers"]["General"]["identifier"]
        amount = dynamoson["headers"]["General"]["amount"]
        amount = re.sub(r"[^\d,.]+", "", amount)
        main.update({
            "identifier": identifier,
            "amount": amount
        })

        dates = dynamoson["tables"]["General Info"][0]["Document Dates"]
        if len(dates) == 0:
            raise Exception(f"At least one date object is required for Lumpers")
        else:
            for d in dates:
                date_dict = get_date_schema()
                date_ = d.get('date_0')
                if date_ is None:
                    raise Exception(f"At least one date is required for Lumpers")
                else:
                    if not bool(re.compile(r"\S").search(date_)):
                        raise Exception(f"Incoming date did not contain valid characters; Date={date_}")
                    elif not bool(re.compile(r"(\d{4})-(\d{2})-(\d{2})").search(date_)):
                        raise Exception(f"Incoming date is incorrect format; Date={date_}")
                    d_ = re.compile(r"(\d{4})-(\d{2})-(\d{2})").search(date_)
                    if d_ is None:
                        raise Exception(f"Incoming date is incorrect format; Date={date_}")
                    else:
                        year = int(d_.group(1))
                        if not bool(re.compile(r"20\d{2}").search(str(year))):
                            raise Exception(f"Year is incorrect; Year={year}")
                        month = int(d_.group(2))
                        if month > 12:
                            raise Exception(f"Month is greater than 12; Month={month}")
                        day = int(d_.group(3))
                        if day > 31:
                            raise Exception(f"Day is greater than 31; Day={day}")

                time = d.get('time_0')
                if time is not None:
                    if bool(re.compile(r"\S").search(time)):
                        time_success, time, time_err = format_time(time)
                        if not time_success:
                            raise Exception(f"format_time() >>> {time_err}")
                        time = time.strftime('%H:%M')
                    else:
                        time = None
                date_dict.update({
                    'date': date_,
                    'time': time,
                    'datetype': 'lumper'
                })
                main['dates'].append(date_dict)

        references = dynamoson["tables"]["General Info"][0]["Document References"]
        if len(references) == 0:
            print(f"No incoming lumper references; skipping...")
        else:
            for r in references:
                idtype = r.get("idtype")
                ref_dict = get_reference_schema(idtype=idtype)
                ref_dict.update({
                    "id": r.get("id")
                })
                main['references'].append(ref_dict)

        print(f"TRANSLATED ASSESSORIAL = {json.dumps(main)}")
        return True, main, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, {}, f"{e}"



import traceback
import datetime
import re


REFERENCE_OPTIONS = {
            "ACCOUNT NUMBER": "4F",
            "APPOINTMENT NUMBER": "AO",
            "BILL OF LADING": "BM",
            "BOOKING NUMBER": "BN",
            "CARRIER REFERENCE": "CR",
            "CHASSIS NUMBER": "KL",
            "CONSIGNEE'S ORDER NUMBER": "CG",
            "CONTAINER NUMBER": "OC",
            "CONTRACT REFERENCE NUMBER": "KL",
            "CUSTOMER'S ORDER NUMBER": "CO",
            "CUSTOMER'S REFERENCE NUMBER": "CR",
            "DELIVERY ORDER NUMBER": "DO",
            "DELIVERY REFERENCE": "KK",
            "DISPATCH PHONE NUMBER": "42",
            "DRIVER PHONE NUMBER": "42",
            "EQUIPMENT NUMBER": "EQ",
            "ID NUMBER": "KL",
            "LOAD PLANNING NUMBER": "LO",
            "MANIFEST NUMBER": "Y5",
            "MC NUMBER": "MCI",
            "MUTUALLY DEFINED": "ZZ",
            "PICKUP APPOINTMENT NUMBER": "PUA",
            "PICKUP REFERENCE NUMBER": "P8",
            "PURCHASE ORDER NUMBER": "PO",
            "RATE CODE NUMBER": "RB",
            "REFERENCE NUMBER": "KL",
            "SALES ALLOWANCE NUMBER": "OT",
            "SERVICE CHARGE NUMBER": "SV",
            "SHIPMENT DESTINATION CODE": "4C",
            "SHIPMENT ORIGIN CODE": "4B",
            "SHIPPER'S IDENTIFYING NUMBER": "SI",
            "SHIPPER'S ORDER NUMBER": "SO",
            "SUBCONTRACT NUMBER": "3X",
            "TRANSACTION REFERENCE NUMBER": "TN",
            "VESSEL NUMBER": "WU"
        }


def get_idtype(_idtype):
    """
    Pass in the id type description to get the id type code.
    """
    func = get_idtype.__name__
    try:
        _idtype = _idtype.upper()
        for k, v in REFERENCE_OPTIONS.items():
            if k == _idtype:
                return v

        raise Exception(f"Failed to find match for {_idtype}")
    except Exception as e:
        print(f"[WARN] {func} Error ==> {e}")
        traceback.print_exc()
        return "KL"


STATES = {
    'alabama': 'al',
    'alaska': 'ak',
    'arizona': 'az',
    'arkansas': 'ar',
    'california': 'ca',
    'colorado': 'co',
    'connecticut': 'ct',
    'delaware': 'de',
    'florida': 'fl',
    'georgia': 'ga',
    'hawaii': 'hi',
    'idaho': 'id',
    'illinois': 'il',
    'indiana': 'in',
    'iowa': 'ia',
    'kansas': 'ks',
    'kentucky': 'ky',
    'louisiana': 'la',
    'maine': 'me',
    'maryland': 'md',
    'massachusetts': 'ma',
    'michigan': 'mi',
    'minnesota': 'mn',
    'mississippi': 'ms',
    'missouri': 'mo',
    'montana': 'mt',
    'nebraska': 'ne',
    'nevada': 'nv',
    'new hampshire': 'nh',
    'new jersey': 'nj',
    'new mexico': 'nm',
    'new york': 'ny',
    'north carolina': 'nc',
    'north dakota': 'nd',
    'ohio': 'oh',
    'oklahoma': 'ok',
    'oregon': 'or',
    'pennsylvania': 'pa',
    'rhode island': 'ri',
    'south carolina': 'sc',
    'south dakota': 'sd',
    'tennessee': 'tn',
    'texas': 'tx',
    'utah': 'ut',
    'vermont': 'vt',
    'virginia': 'va',
    'washington': 'wa',
    'west Virginia': 'wv',
    'wisconsin': 'wi',
    'wyoming': 'wy',
    'district of columbia': 'dc'
}


def format_state(state: str, abbreviate: bool):
    """
    Checks and normalizes a US state name according to the <abbreviate> parameter.
    :param state: str,;
    :param abbreviate: bool
    """
    func = format_state.__name__
    try:
        found = False
        state = re.sub(r" \B", "", state.lower())  # remove trailing end space
        state = re.sub(r"\B ", "", state.lower())  # remove trailing start space
        state = re.sub(r" +", " ", state.lower())  # remove extra spaces
        val = None
        for s, abbrev in STATES.items():
            if s == state:
                found = True
                if abbreviate:
                    val = abbrev.upper()
                    break
                else:
                    val = s.title()
                    break

            elif abbrev == state:
                found = True
                if abbreviate:
                    val = abbrev.upper()
                    break
                else:
                    val = s.title()
                    break

        if not found:
            raise Exception(f"State ({state}) is not valid")

        return found, val, f"Abbreviated: {abbreviate}"
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"


def parse_incoming_address(address):
    """
    Parses an incoming address from the DT to a normalized string;
    Accepts: strings, list of strings, list-of-list of strings
    :param address: <who knows>, incoming address from the DT;
    :return: bool(success), str(address), str(error).
    """
    func = parse_incoming_address.__name__
    try:
        if isinstance(address, list):
            if len(address) == 0:
                return True, "", ""
            if len(address) == 1:
                address = address[0]
                if isinstance(address, str):
                    address = address.title()
                elif isinstance(address, list):
                    if len(address) == 1:
                        address = address[0].title()
                    else:
                        address = [a.title() for a in address]
                        address = ", ".join(address)
            else:
                if isinstance(address[0], str):
                    address = [a.title() for a in address]
                    address = ", ".join(address)
                else:
                    address_ = list()
                    for a_list in address:
                        if isinstance(a_list[0], str):
                            address_.extend([a.title() for a in a_list])
                    address = ", ".join(address_)

        print(f"[{func}] DataType: {type(address)}; Value: {address}")

        return True, address, ""
    except Exception as e:
        print(f"[ERROR] {func} ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"


def parse_time(time_str):
    func = parse_time.__name__
    try:
        try:
            time_ = datetime.datetime.strptime(time_str, '%I:%M %p')
            time_str = time_.strftime('%H:%M')
        except ValueError as TimeError1:
            print(f"[WARN] TimeError1 ==> {TimeError1}")
            try:
                time_ = datetime.datetime.strptime(time_str, '%I:%M%p')
                time_str = time_.strftime('%H:%M')
            except ValueError as TimeError2:
                print(f"[WARN] TimeError2 ==> {TimeError2}")
                try:
                    time_ = datetime.datetime.strptime(time_str, '%H:%M:%S.%f')
                    time_str = time_.strftime('%H:%M')
                except ValueError as TimeError3:
                    print(f"[WARN] TimeError3 ==> {TimeError3}")
                    try:
                        time_ = datetime.datetime.strptime(time_str, '%H:%M:%S')
                        time_str = time_.strftime('%H:%M')
                    except ValueError as TimeError4:
                        print(f"[WARN] TimeError4 ==> {TimeError4}")
                        try:
                            time_ = datetime.datetime.strptime(time_str, '%H:%M')
                            time_str = time_.strftime('%H:%M')
                        except ValueError as TimeError5:
                            print(f"[WARN] TimeError5 ==> {TimeError5}")
                            raise Exception(f"No time-patterns matched: {time_str}")

        return True, time_str, ""
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"


def format_time(time):
    func = format_time.__name__
    try:
        try:
            time_ = datetime.datetime.strptime(time, '%I:%M %p')
        except Exception as ee:
            print(f"[{func} Warning] {ee}")
            time_ = datetime.datetime.strptime(time, '%H:%M')

        return True, time_, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"



"""
Custom Exceptions for custom problems.

Helps to isolate some specific issues and customize the way they are handled, rather than leaning on the standard
Exception type or ValueError.

Works the same way that a typical Exception would.
"""


class SelectionOutOfRange(Exception):
    """
    Raised if a page selection is passed in that exceeds the maximum number of pages in the PDF.

    e.i. input = "1-7" but the total pages in the PDF is 5.
    """
    pass


class ReversedRange(Exception):
    """
    Raised in the event that a simple range is passed in where the start is bigger than the end.

    i.e. input = "10-5"

    Page range numbers must be in order from smallest >> largest.
    """
    pass


class UnbalancedPageRange(Exception):
    """
    Raised if the page range (start-end) where 'end' is a number that is higher
    than the numbers that follow it separated by a comma.

    i.e. input = "3-7, 4, 5"

    The numbers that follow are already covered by the preceding range.
    """
    pass


class NotZeroIndex(Exception):
    """
    Raised in the event that an annotator passes in a page range that includes "0" as a page number.
    Document page ranges coming from the DT are not a zero-index.
    """
    pass



import json
import decimal
import traceback
import urllib.parse
from dukeai_lib.globals import CUSTOMER_EMAIL_FOOTER


def send_email(
        subject: str,
        body: str,
        recipient: str,
        email_sender: str,
        footer: bool,
        email_client,
        body2="",
        body3="",
        body4="",
        body5="",
        body6="",
        body7="",
        body8=""
):
    func = send_email.__name__
    if isinstance(recipient, str):
        recipient = [recipient]

    if body2 is None:
        body2 = ""
    if body3 is None:
        body3 = ""
    if body4 is None:
        body4 = ""
    if body5 is None:
        body5 = ""
    if body6 is None:
        body6 = ""
    if body7 is None:
        body7 = ""
    if body8 is None:
        body8 = ""

    if footer:
        body_html = f"""
        <html>
        <head></head>
        <body>
        <p>{body}</p>
        <p>{body2}</p>
        <p>{body3}</p>
        <p>{body4}</p>
        <p>{body5}</p>
        <p>{body6}</p>
        <p>{body7}</p>
        <p>{body8}</p>
        {CUSTOMER_EMAIL_FOOTER}
        </body>
        </html>
        """
    else:
        body_html = f"""
        <html>
        <head></head>
        <body>
        <p>{body}</p>
        <p>{body2}</p>
        <p>{body3}</p>
        <p>{body4}</p>
        <p>{body5}</p>
        <p>{body6}</p>
        <p>{body7}</p>
        <p>{body8}</p>
        </body>
        </html>
        """
    try:
        response = email_client.send_email(
            Destination={
                'ToAddresses': recipient
            },
            Message={
                'Body': {
                    'Html': {
                        'Charset': "UTF-8",
                        'Data': body_html,
                    },
                    'Text': {
                        'Charset': "UTF-8",
                        'Data': (str(body_html)),
                    }
                },
                'Subject': {
                    'Charset': "UTF-8",
                    'Data': subject,
                },
            },
            Source=email_sender
        )
        print(f"Email sent! Message ID: {response['MessageId']}")

        return True, response, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, {}, f"{e}"


def gen_single_button_html(button_text, button_header, link, alignment, button_color, button_text_color):
    link_style1 = f"""
    style="padding: 8px 12px; font-size: 14px; color: {button_text_color}; font-weight:bold; display:inline-block;"
    """
    button = f"""
    <br>
    <br>
    <table width="100%" cellspacing="0" cellpadding="0", border="0", align="{alignment}" style="margin-bottom:50px; max-width:700px; width:100%;">
        <tr>
            <th>{button_header}</th>
        </tr>
        <tr>

            <td style="border-radius: 2px; text-align:center; vertical-align:middle">
                <table cellspacing="0" cellpadding="0", align="center">
                    <tr style="text-align:center; vertical-align:middle">
                        <td class="button" bgcolor="{button_color}" style="font-family:Arial,Helvetica,sans-serif; font-size:14px; font-weight:bold; display:inline-block; background-clip:padding-box;">
                            <a class=”link” href="{link}" target="_blank" {link_style1}>
                                {button_text}             
                            </a>
                        </td>
                    </tr>
                </table>
            </td>

        </tr>
    </table>
    <br>
    <br>
    <br>
    <br>
    <br>
    <br>
    """
    return button


def gen_three_button_html(bd_array: list):
    """
    Generates raw HTML for an inline 3 button email insert;
    Expected input for 'bd_array' is something like:
    bd_array=[
        {
            'button_header': 'Button 1',
            'button_text': 'option 1',
            'link': "https://duke.ai/accountant-portal/signin/index.html"
        },
        {
            'button_header': 'Button 2',
            'button_text': 'option 2',
            'link': "https://duke.ai/accountant-portal/signin/index.html"
        },
        {
            'button_header': 'Button 3',
            'button_text': 'option 3',
            'link': "https://duke.ai/accountant-portal/signin/index.html"
        }
    ]
    """
    func = gen_three_button_html.__name__
    try:
        link_style1 = """
        style="padding: 8px 12px; font-size: 14px; color: #ffffff; font-weight:bold; display:inline-block;"
        """
        table = f"""
        <table width="100%" cellspacing="4" cellpadding="4", border="0", align="center" style="margin-bottom:50px; max-width:700px; width:100%;">
            <tr>
                <th>{bd_array[0]['button_header']}</th>
                <th>{bd_array[1]['button_header']}</th>
                <th>{bd_array[2]['button_header']}</th>
            </tr>
            <tr>

                <td style="border-radius: 2px; text-align:center; vertical-align:middle">
                    <table cellspacing="0" cellpadding="0", align="center">
                        <tr style="text-align:center; vertical-align:middle">
                            <td class="button" bgcolor="#1F83FF" style="font-family:Arial,Helvetica,sans-serif; font-size:14px; font-weight:bold; display:inline-block; background-clip:padding-box;">
                                <a class=”link” href="{bd_array[0]['link']}" target="_blank" {link_style1}>
                                    {bd_array[0]['button_text']}             
                                </a>
                            </td>
                        </tr>
                    </table>
                </td>

                <td style="border-radius: 2px; text-align:center; vertical-align:middle">
                    <table cellspacing="0" cellpadding="0", align="center">
                        <tr style="text-align:center; vertical-align:middle">
                            <td class="button" bgcolor="#1F83FF" style="font-family:Arial,Helvetica,sans-serif; font-size:14px; font-weight:bold; display:inline-block; background-clip:padding-box;">
                                <a class=”link” href="{bd_array[1]['link']}" target="_blank" {link_style1}>
                                    {bd_array[1]['button_text']}             
                                </a>
                            </td>
                        </tr>
                    </table>
                </td>

                <td style="border-radius: 2px; text-align:center; vertical-align:middle">
                    <table cellspacing="0" cellpadding="0", align="center">
                        <tr style="text-align:center; vertical-align:middle">
                            <td class="button" bgcolor="#1F83FF" style="font-family:Arial,Helvetica,sans-serif; font-size:14px; font-weight:bold; display:inline-block; background-clip:padding-box;">
                                <a class=”link” href="{bd_array[2]['link']}" target="_blank" {link_style1}>
                                    {bd_array[2]['button_text']}             
                                </a>
                            </td>
                        </tr>
                    </table>
                </td>

            </tr>
        </table>
        """

        return True, table
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, ""


def format_bucket_url(bucket_name: str, full_object_name: str):
    func = format_bucket_url.__name__
    try:
        full_object_name = urllib.parse.quote(full_object_name)
        response = f"https://{bucket_name}.s3.amazonaws.com/{full_object_name}"
        print(f"[{func}] Returning {response}")
        return response
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return None


def multi_load_confirmation_email(
        subject: str,
        body: str,
        recipient: str,
        invoice_response_data: list,
        bol_response_data: list,
        email_client,
        email_sender,
        body2="",
        body3=""
):
    function_name = multi_load_confirmation_email.__name__
    if isinstance(recipient, str):
        recipient = [recipient]

    bol_responses = list()
    for bol_res in bol_response_data:
        idx = bol_response_data.index(bol_res)
        button = gen_single_button_html(
            button_text="Download",
            button_header="BOL Load documents",
            link=bol_res['load_document_link'],
            alignment="center",
            button_color="#1F83FF",
            button_text_color="#ffffff"
        )
        bol_responses.append(
            f"""
            <p>Bill of Lading Reference #: {bol_res['kvt_data']['identifier']}</p>
            <p>Total: ${bol_res['kvt_data']['charges']}</p>
            {button}
            """
        )
        if idx != (len(bol_responses) - 1):
            bol_responses.append(
                "<p>-----------------------------------</p>"
            )
        else:
            if len(invoice_response_data) >= 1:
                bol_responses.append(
                    "<p>-----------------------------------</p>"
                )
            else:
                bol_responses.append("<p></p>")

    load_responses = list()
    load_responses.extend(bol_responses)
    for load_resp in invoice_response_data:
        idx = invoice_response_data.index(load_resp)

        button_data = [
            {
                'button_header': 'Invoice',
                'button_text': 'Download PDF',
                'link': load_resp['duke_invoice_url']
            },
            {
                'button_header': 'DUKE Verified Score',
                'button_text': 'Download PDF',
                'link': load_resp['score_report_url']
            },
            {
                'button_header': 'Invoiced Load Documents',
                'button_text': 'Download PDF',
                'link': load_resp['load_document_link']
            }
        ]
        success, table = gen_three_button_html(button_data)
        if not success:
            raise Exception(f"gen_three_button_html encountered an error")

        if load_resp['vendor_bill_no'] not in ["0000", "00", "0"]:
            load_responses.append(
                f"""
                <p>Rate Confirmation Reference #: {load_resp['load_reference']}</p>
                <p>DUKE.ai Invoice #: {load_resp['duke_invoice_no']}</p>
                <p>Vendor Bill #: {load_resp['vendor_bill_no']}</p>
                <p>DUKE.ai Load reference #: {load_resp['load_id']}</p>
                <p>Total: ${load_resp['amount']}</p>
                {table}
                """
            )

        else:
            # """
            # <p>Downloadable Invoice PDF: {load_resp['duke_invoice_url']}</p>
            # <p>Downloadable Load Acceptance Score PDF: {load_resp['score_report_url']}</p>
            # <p>Downloadable Load Document PDF: {load_resp['load_document_link']}</p>
            # """
            load_responses.append(
                f"""
                <p>Rate Confirmation Reference #: {load_resp['load_reference']}</p>
                <p>DUKE.ai Invoice #: {load_resp['duke_invoice_no']}</p>
                <p>DUKE.ai Load reference #: {load_resp['load_id']}</p>
                <p>Total: ${load_resp['amount']}</p>
                {table}
                """
            )

        if idx != (len(invoice_response_data) - 1):
            load_responses.append(
                "<p>-----------------------------------</p>"
            )
        else:
            load_responses.append("<p></p>")

    load_responses.append(
        f"""
        <p></p>
        <p>-----------------------------------</p>
        <p></p>
        <p>Please contact Support@Duke.ai if you have any questions, and thank you for trucking with DUKE.ai!</p>
        """
    )

    body_string = "".join(load_responses)

    body_html = f"""
    <html>
    <head></head>
    <body>
    <p>{body}</p>
    <p>{body2}</p>
    <p>{body3}</p>
    <p></p>
    {body_string}
    </body>
    </html>
      """
    try:
        response = email_client.send_email(
            Destination={
                'ToAddresses': recipient
            },
            Message={
                'Body': {
                    'Html': {
                        'Charset': "UTF-8",
                        'Data': body_html,
                    },
                    'Text': {
                        'Charset': "UTF-8",
                        'Data': (str(body_html)),
                    }
                },
                'Subject': {
                    'Charset': "UTF-8",
                    'Data': subject,
                },
            },
            Source=email_sender
        )
        print(f"[INFO] {function_name} Email sent! Message ID: {response}")
        return response
    except Exception as e:
        print(f"[ERROR] {function_name} Error ==> {e}")
        traceback.print_exc()
        return None


def send_schedule_email(
        subject,
        body,
        recipient,
        duke_customer,
        pdf_url,
        platform_vendor_id,
        transaction_fee,
        factoring_fee,
        all_response_data,
        email_sender,
        email_client,
        zoho=False,
        quickbooks=False
):
    func = send_schedule_email.__name__
    try:
        if isinstance(recipient, str):
            recipient = [recipient]

        load_responses = list()
        for load_resp in all_response_data:
            idx = all_response_data.index(load_resp)
            report_button = gen_single_button_html(
                button_header=f"DUKE Verified Score - {load_resp['load_reference']}",
                button_text="Download PDF",
                link=load_resp['score_report_url'],
                alignment="left",
                button_color="#1F83FF",
                button_text_color="#ffffff"
            )
            load_responses.append(
                f"""
                <p>Your bill number: {load_resp['vendor_bill_no']}</p>
                <p>Carrier's DUKE Invoice number: {load_resp['duke_invoice_no']}</p>
                <p>Load reference number: {load_resp['load_reference']}</p>
                <p>Rate Confirmation amount: {load_resp['amount']}</p>
                <p>{report_button}</p>
                """
            )
            if idx != (len(all_response_data) - 1):
                load_responses.append(
                    "<p>-----------------------------------</p>"
                )
            else:
                load_responses.append("<p></p>")

        platform = ""
        if zoho and not quickbooks:
            platform = "Zoho"
        elif quickbooks and not zoho:
            platform = "Quickbooks"
        elif not zoho and not quickbooks:
            raise Exception(f"Both values of <zoho> and <quickbooks> are set to False! One must be True.")

        schedule_button = gen_single_button_html(
            button_header="Schedule of Invoices",
            button_text="Download PDF",
            link=pdf_url,
            alignment="left",
            button_color="#1F83FF",
            button_text_color="#ffffff"
        )

        body_string = "".join(load_responses)
        body_text = str(body)
        body_html = f"""
        <html>
        <head>Schedule of Invoices Purchased from {duke_customer.upper()}</head>
        <body>
        <p>{body_text}</p>
        <p></p>
        <p>For your convenience, please note: the total amount paid to your client on this schedule includes a transaction fee deduction of {transaction_fee}, as well as a factoring fee deduction of {factoring_fee}% as specified in your account with DUKE.ai. 
        The bill that exists in your {platform} account does not include these deductions, and will need to be updated at the time of payment. Any custom line items that you add, or changes to the data or amount will reflect on the Carrier's
        Invoice in the DUKE system within 24 hours.</p>
        <p></p>
        <p>DUKE customer email address: {duke_customer.upper()}</p>
        <p>{platform} vendor ID for customer: {platform_vendor_id}</p>
        <p>{schedule_button}</p>
        <p></p>
        <p>-----------------------------------</p>
        <p></p>
        {body_string}
        <p>Thanks for trucking with DUKE.ai!</p>
        <p>Support@DUKE.ai</p>
        <p>831.350.4529</p>
        </body>
        </html>
        """

        response = email_client.send_email(
            Destination={
                'ToAddresses': recipient
            },
            Message={
                'Body': {
                    'Html': {
                        'Charset': "UTF-8",
                        'Data': body_html,
                    },
                    'Text': {
                        'Charset': "UTF-8",
                        'Data': str(body_text),
                    },
                },
                'Subject': {
                    'Charset': "UTF-8",
                    'Data': subject,
                },
            },
            Source=email_sender
        )
        print(f"[INFO] Email sent! Message ID: {response}")
        return True, response, ""
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, ""


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            if o % 1 > 0:
                return float(o)
            else:
                return float(o)
        return super(DecimalEncoder, self).default(o)



# noa_lor_flattener.py
import json
import traceback
from dukeai_lib.utilities import DecimalEncoder


def get_main_schema():
    main_schema = {
        "headers": {
            "General": {},
            "NOA Information": {},
            "LOR Information": {}
        },
        "tables": {}
    }
    return main_schema


def get_general_schema():
    schema = {
        "broker_id": "",
        "carrier_name": "",
        "carrier_mc_no": "",
        "date": "",
        "broker_name": ""
    }
    return schema


def get_noa_schema():
    schema = {
        "noa_factoring_company_id": "",
        "noa_factoring_company_name": "",
        "noa_page_number": ""
    }
    return schema


def get_lor_schema():
    schema = {
        "lor_factoring_company_id": "",
        "lor_factoring_company_name": "",
        "lor_page_number": ""
    }
    return schema


def fail_schema():
    main = dict(get_main_schema())
    main["headers"].update({
        "General": dict(get_general_schema()),
        "NOA Information": dict(get_noa_schema()),
        "LOR Information": dict(get_lor_schema())
    })
    return main


def flatten_noa_lor(ddb_obj: dict) -> tuple[bool, dict, str]:
    """
    Takes a DDB object and flattens data into Dynamotator schema;
    :param ddb_obj: dict, DUKE-User-Document object;
    :return: str, flattened schema for Dynamotator UI.
    """
    # known special characters to remove that don't play nicely with JSON
    func = flatten_noa_lor.__name__
    char_dict = r"\"|\'|\x0c"
    print(f"[INFO] {func} activated")
    try:
        print(f"[INCOMING DDB OBJECT] {ddb_obj}")
        ddb_copy = ddb_obj.copy()
        auto_ex = ddb_obj.get("autoextracts")

        if auto_ex is None:
            print(f"[WARN] DDB entry has no autoextracts; Defaulting to empty schema")
            ddb_copy["autoextracts"]["data"] = fail_schema()
            return True, ddb_copy, ""

        else:
            data = auto_ex.get("data")
            if data is None or (isinstance(data, dict) and len(data.keys()) == 0):
                print(f"[WARN] autoextracts has no data object; Defaulting to empty schema")
                ddb_copy["autoextracts"]["data"] = fail_schema()
                return True, ddb_copy, ""

            else:
                empty = True
                for k, v in data.items():
                    if v is not None:
                        empty = False
                if empty:
                    print(f"[WARN] No data found in Autoextracts, checking Extracts....")
                    data = ddb_obj.get("extracts")
                    if isinstance(data, dict):
                        if 'data' in data.keys():
                            data = data['data']
                            for k, v in data.items():
                                if v is not None:
                                    empty = False
                if empty:
                    print(f"[WARN] Data is empty in Extracts AND Autoextracts")
                    ddb_copy["autoextracts"]["data"] = fail_schema()
                    return True, ddb_copy, "Data is empty in Extracts AND Autoextracts"

                main = get_main_schema()
                main_headers = dict(main["headers"])
                general = get_general_schema()
                date = data.get("date")
                cname = data.get("carrier_name")
                cmc = data.get("carrier_mc_no")
                bid = data.get("broker_id")
                bname = data.get("broker_name")
                general.update({
                    "broker_id": bid,
                    "carrier_name": cname,
                    "carrier_mc_no": cmc,
                    "date": date,
                    "broker_name": bname
                })
                for k, v in general.items():
                    if v is None:
                        general.update({k: ""})
                    else:
                        pass
                main_headers.update({"General": general})
                print(f"[DEBUG] general = {general}")

                noa = get_noa_schema()
                nfc_id = data.get("noa_factoring_company_id")
                nfc_name = data.get("noa_factoring_company_name")
                n_pn = data.get("noa_page_number")
                noa.update({
                    "noa_factoring_company_id": nfc_id,
                    "noa_factoring_company_name": nfc_name,
                    "noa_page_number": n_pn
                })
                for k, v in noa.items():
                    if v is None:
                        noa.update({k: ""})
                    else:
                        pass
                main_headers.update({"NOA Information": noa})
                print(f"[DEBUG] noa = {noa}")

                lor = get_lor_schema()
                lfc_id = data.get("lor_factoring_company_id")
                lfc_name = data.get("lor_factoring_company_name")
                l_pn = data.get("lor_page_number")
                lor.update({
                    "lor_factoring_company_id": lfc_id,
                    "lor_factoring_company_name": lfc_name,
                    "lor_page_number": l_pn
                })
                for k, v in lor.items():
                    if v is None:
                        lor.update({k: ""})
                    else:
                        pass
                main_headers.update({"LOR Information": lor})
                print(f"[DEBUG] lor = {lor}")

                main["headers"] = main_headers

                try:
                    flattened_data = json.loads(json.dumps(main, indent=2, cls=DecimalEncoder, separators=(',', ': ')))
                    ddb_copy["autoextracts"]["data"] = flattened_data
                    print("[INFO] flatten_noa_lor() completed")
                    return True, ddb_copy, ""
                except Exception as ee:
                    print(f"[FAIL] Dictionary contains characters that are illegal JSON syntax; Defaulting to blank schema; ERROR = {ee}")
                    ddb_copy["autoextracts"]["data"] = fail_schema()
                    return True, ddb_copy, f"{ee}"

    except Exception as e:
        print(f"[ERROR] {func} failed; Error ==> {e}")
        traceback.print_exc()
        try:
            ddb_copy = ddb_obj.copy()
            ddb_copy["autoextracts"]["data"] = fail_schema()
            return False, ddb_copy, f"{e}"
        except Exception as eee:
            print(f"[DOUBLE FAIL] flatten_data for (noa_lor) failed to generate blank schema to return; ERROR = {eee}")
            traceback.print_exc()
            return False, {}, f"{eee}"


def get_noa_lor_schema():
    schema = {
        "date": None,
        "carrier_name": None,
        "carrier_mc_no": None,
        "noa_factoring_company_name": None,
        "lor_factoring_company_name": None,
        "noa_page_number": None,
        "lor_page_number": None
    }
    return schema


def unflatten_noa_lor(flattened_ddb_entry: str) -> tuple[bool, dict, str]:
    """
    Takes flattened data from the Dynamotator UI and rebuilds the information into the original Schema;
    :param flattened_ddb_entry: JSON str;
    :return: JSON str, Original ratecon schema inside the ddb_entry.
    """
    func = unflatten_noa_lor.__name__
    if isinstance(flattened_ddb_entry, dict):
        flattened_ddb_entry = json.dumps(flattened_ddb_entry)

    data = json.loads(flattened_ddb_entry)
    # data = data['autoextracts']['data']

    main = get_noa_lor_schema()
    try:
        general = data["headers"].get("General")
        date = general.get("date")
        carrier_name = general.get("carrier_name")
        carrier_mc_no = general.get("carrier_mc_no")
        main.update({
            "date": date,
            "carrier_name": carrier_name,
            "carrier_mc_no": carrier_mc_no
        })

        noa = data["headers"].get("NOA Information")
        noa_factoring_company_name = noa.get("noa_factoring_company_name")
        noa_page_number = noa.get("noa_page_number")
        main.update({
            "noa_factoring_company_name": noa_factoring_company_name,
            "noa_page_number": noa_page_number
        })

        lor = data["headers"].get("LOR Information")
        lor_factoring_company_name = lor.get("lor_factoring_company_name")
        lor_page_number = lor.get("lor_page_number")
        main.update({
            "lor_factoring_company_name": lor_factoring_company_name,
            "lor_page_number": lor_page_number
        })

        return True, main, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, {}, f"{e}"



import secrets
import base58
import hashlib
import datetime


def gen_random_sha(length=None):
    """
    Generate a random sha256 string;
    :param length: int, the integer argument to pass to secrets.token_urlsafe()
    :return: str, sha256.
    """
    if length is None:
        length = 36
    milli_date = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S.%f")
    random_num = f"{secrets.token_urlsafe(length)}:{milli_date}"
    sha256 = base58.b58encode(hashlib.sha256(random_num.encode()).digest())
    string_hash = ''.join(map(chr, sha256))
    return string_hash



"""

dukeai_lib

Common functions used across the DUKE.ai project environments.

"""

from . import application, tools, utilities, schema_kung_fu, load_transmission, globals

__version__ = "0.2.2"
__author__ = "Blake Donahoo"






import re
import json
import traceback
from .schema_utilities import get_idtype, format_time, format_state


def get_flat_entity_schema():
    return {
        "type": None,
        "name": None,
        "address": None,
        "city": None,
        "state": None,
        "postal": None,
        "country": None,
        "contactname": None,
        "contact_number": None
    }


def get_flat_date_schema():
    return {
        "date_0": None,
        "datetype_0": None,
        "time_0": None
    }


def get_flat_reference_schema():
    return {
        "id": None,
        "idtype": None
    }


def get_flat_base_schema():
    return {
        "headers": {
            "General": {
                "identifier": None,
                "delivery_signature": False,
                "is_delivered": False,
                "charges": None
            }
        },
        "tables": {
            "General Info": [
                {
                    "Document Dates": [],
                    "Document References": [],
                    "Document Entities": []
                }
            ]
        }
    }


def blank_schema():
    blank = get_flat_base_schema()
    consignee_entity = get_flat_entity_schema()
    consignee_entity.update({'type': 'CONSIGNEE'})
    shipper_entity = get_flat_entity_schema()
    shipper_entity.update({'type': 'SHIPPER'})
    carrier_entity = get_flat_entity_schema()
    carrier_entity.update({'type': 'CARRIER'})

    shipped_date = get_flat_date_schema()
    shipped_date.update({'datetype_0': 'General'})
    delivered_date = get_flat_date_schema()
    delivered_date.update({'datetype_0': 'Delivered'})

    blank['tables']['General Info'][0].update({
        "Document Dates": [shipped_date, delivered_date],
        "Document References": [get_flat_reference_schema()],
        "Document Entities": [consignee_entity, shipper_entity, carrier_entity]
    })
    return blank


def flatten_bol(ddb_entry) -> tuple[bool, dict, str]:
    """
    Translates the BOL KVT into Dynamoson.
    """
    func = flatten_bol.__name__
    if isinstance(ddb_entry, str):
        ddb_entry = json.loads(ddb_entry)
    ddb_entry = ddb_entry.copy()
    try:
        send_blank = False
        auto_ex = ddb_entry.get('autoextracts')

        if auto_ex is None:
            print(f"[WARN] 'ddb_entry' has no 'autoextracts' object")
            send_blank = True

        elif auto_ex.get('data') is None:
            print(f"[WARN] 'ddb_entry'[autoextracts] has no 'data' object")
            send_blank = True

        elif len(auto_ex['data'].keys()) == 0:
            print(f"[WARN] 'ddb_entry'[autoextracts][data] has a key-length of 0")
            send_blank = True

        if send_blank:
            blank = blank_schema()
            ddb_entry.update({
                "autoextracts": {
                    "data": blank
                }
            })
            return True, ddb_entry, ""

        ### Begin mapping data from the KVT to the Dynamoson
        data = auto_ex["data"]
        main = get_flat_base_schema()

        identifier = str(data['identifier'])
        del_sig = data["delivery_signature"]
        if isinstance(del_sig, str):
            if del_sig.lower() == "true":
                del_sig = True
            elif del_sig.lower() == "false":
                del_sig = False
        is_delivered = data["is_delivered"]
        if isinstance(is_delivered, str):
            if is_delivered.lower() == "true":
                is_delivered = True
            elif is_delivered.lower() == "false":
                is_delivered = False
        charges = str(data["charges"])

        main["headers"]["General"].update({
            "identifier": identifier,
            "delivery_signature": del_sig,
            "is_delivered": is_delivered,
            "charges": charges
        })

        dates = data["dates"]
        shipped_date = False
        delivered_date = False
        for d in dates:
            date_dict = get_flat_date_schema()

            if not bool(re.compile(r"\S").search(d['date'])):
                print("Skipping blank date dict...")
                continue

            if d.get('datetype') is not None:
                datetype = d['datetype']
                if datetype.lower() in ["shipped", "general"]:
                    shipped_date = True
                elif datetype.lower() == "delivered":
                    delivered_date = True

            date_dict.update({
                "date_0": d["date"],
                "datetype_0": d["datetype"],
                "time_0": d["time"]
            })
            main["tables"]["General Info"][0]["Document Dates"].append(date_dict)
        if not shipped_date:
            shipped_date_ = get_bol_date_schema()
            shipped_date_.update({'datetype_0': 'General'})
            main["tables"]["General Info"][0]["Document Dates"].append(shipped_date_)
        if not delivered_date:
            delivered_date_ = get_bol_date_schema()
            delivered_date_.update({'datetype_0': 'Delivered'})
            main["tables"]["General Info"][0]["Document Dates"].append(delivered_date_)

        references = data["references"]
        for r in references:

            if not bool(re.compile(r"\S").search(r['id'])):
                print("Skipping blank date dict...")
                continue

            ref = get_flat_reference_schema()
            ref.update({
                "id": r["id"],
                "idtype": r["idtype"]
            })
            main["tables"]["General Info"][0]["Document References"].append(ref)
        if len(main["tables"]["General Info"][0]["Document References"]) == 0:
            ref_ = get_flat_reference_schema()
            main["tables"]["General Info"][0]["Document References"].append(ref_)

        entities = data["entities"]
        consignee = False
        shipper = False
        carrier = False
        for e in entities:
            ent = get_flat_entity_schema()

            if e.get('type') is not None:
                ent_type = e['type']
                if ent_type.lower() == "consignee":
                    consignee = True
                elif ent_type.lower() == "shipper":
                    shipper = True
                elif ent_type.lower() == "carrier":
                    carrier = True

            ent.update({
                "name": e["name"],
                "type": e["type"],
                "city": e["city"],
                "state": e["state"],
                "postal": e["postal"],
                "country": e["country"],
                "contactname": e["contacts"]["contactname"],
                "contact_type": e["contacts"]["contact_type"],
                "contact_number": e["contacts"]["contact_number"],
                "contact_number_type": e["contacts"]["contact_number_type"]
            })
            address = None
            address_ = e["address"]
            if isinstance(address_, str):
                address = address_.upper()
            elif isinstance(address_, list):
                address = list()
                for a in address_:
                    if isinstance(a, str):
                        address.append(a.upper())
                    elif isinstance(a, list):
                        for aa in a:
                            address.append(aa.upper())
                address = " ".join(address)

            ent["address"] = address
            main["tables"]["General Info"][0]["Document Entities"].append(ent)

        if not shipper:
            shipper_ = get_flat_entity_schema()
            shipper_.update({'type': 'SHIPPER'})
            main["tables"]["General Info"][0]["Document Entities"].append(shipper_)
        if not consignee:
            consignee_ = get_flat_entity_schema()
            consignee_.update({'type': 'CONSIGNEE'})
            main["tables"]["General Info"][0]["Document Entities"].append(consignee_)
        if not carrier:
            carrier_ = get_flat_entity_schema()
            carrier_.update({'type': 'CARRIER'})
            main["tables"]["General Info"][0]["Document Entities"].append(carrier_)

        try:
            flattend_data = json.loads(json.dumps(main))
            ddb_entry['autoextracts']['data'] = flattend_data
            print(f"[FLATTENED DATA] {flattend_data}")
            return True, ddb_entry, ""
        except Exception as ee:
            print(f"[WARN] Could not dump and load flattened data ==> {ee}")
            blank = blank_schema()
            ddb_entry.update({
                "autoextracts": {
                    "data": blank
                }
            })
            return True, ddb_entry, f"{ee}"

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        blank = blank_schema()
        ddb_entry.update({
            "autoextracts": {
                "data": blank
            }
        })
        return False, ddb_entry, f"{e}"


def get_bol_reference_schema(idtype):
    _idtype = get_idtype(idtype)
    return {
        "id": None,
        "idtype": idtype,
        "_idtype": _idtype
    }


def get_bol_date_schema():
    return {
        "date": None,
        "datetype": None,
        "time": None
    }


def get_bol_entity_schema(type_):
    func = get_bol_entity_schema.__name__
    _type = None
    if type_.upper() == "SHIPPER":
        _type = "SH"
    elif type_.upper() == "CONSIGNEE":
        _type = "CN"
    elif type_.upper() == "CARRIER":
        _type = "CR"
    elif type_.upper() == "BILL TO":
        _type = "BT"
    elif type_.upper() == "BROKER":
        _type = "BK"
    else:
        print(f"[WARN][{func}] {type_} unable to be matched to type-code.")
    return {
        "name": None,
        "type": type_.upper(),
        "_type": _type,
        "city": None,
        "state": None,
        "postal": None,
        "country": None,
        "address": [],
        "contacts": {
            "contactname": None,
            "contact_type": None,
            "contact_number": None,
            "contact_number_type": None
        }
    }


def get_bol_schema():
    return {
        "identifier": None,
        "delivery_signature": False,
        "is_delivered": False,
        "charges": None,
        "dates": [],
        "references": [],
        "entities": []
    }


def unflatten_bol(dynamoson) -> tuple[bool, dict, str]:
    """
    Translate flattened BOL data back into the KVT schema
    """
    func = unflatten_bol.__name__
    if isinstance(dynamoson, str):
        dynamoson = json.loads(dynamoson)

    try:
        main = get_bol_schema()
        general = dynamoson["headers"]["General"]
        identifier = general["identifier"]
        delivery_signature = general["delivery_signature"]
        if isinstance(delivery_signature, str):
            if delivery_signature.lower() in ["true", "yes", "y"]:
                delivery_signature = True
            elif delivery_signature.lower() in ["false", "no", "n"]:
                delivery_signature = False
        is_delivered = general["is_delivered"]
        if isinstance(is_delivered, str):
            if is_delivered.lower() in ["true", "yes", "y"]:
                is_delivered = True
            elif is_delivered.lower() in ["false", "no", "n"]:
                is_delivered = False
        charges = str(general["charges"])
        charges = re.sub(r"[^\d.]+", "", charges)

        main.update({
            "identifier": identifier,
            "delivery_signature": delivery_signature,
            "is_delivered": is_delivered,
            "charges": charges
        })

        dates = dynamoson["tables"]["General Info"][0]["Document Dates"]
        shipped_date = False
        delivered_date = False
        for d in dates:
            if d['date_0'] is None:
                continue
            if not bool(re.compile(r"\S+").search(d['date_0'])):
                continue
            if not bool(re.compile(r"\d{4}-\d{2}-\d{2}").search(d['date_0'])):
                raise Exception(f"Date {d['date_0']} is improperly formatted")

            date_dict = get_bol_date_schema()
            date = d["date_0"]
            time = d["time_0"]
            if time is not None:
                if bool(re.compile(r"\S").search(time)):
                    time_success, time, time_err = format_time(time)
                    if not time_success:
                        raise Exception(f"format_time() >>> {time_err}")
                    time = time.strftime('%H:%M')
                else:
                    time = None
            else:
                time = None

            if d.get("datetype_0") is None:
                raise Exception(f"{d['date_0']} had a datetype of {d['datetype_0']} submitted. A valid datetype must be selected.")
            if d["datetype_0"].upper() in ["GENERAL", "SHIPPED"]:
                shipped_date = True
            elif d["datetype_0"].upper() == "DELIVERED":
                delivered_date = True

            date_dict.update({
                "date": date,
                "datetype": d["datetype_0"].upper(),
                "time": time
            })
            main["dates"].append(date_dict)

        if not shipped_date:
            print(f"[WARN] No shipping date found!")
            return False, {}, "Shipment or General Document date required"

        elif not delivered_date and (is_delivered is True or delivery_signature is True):
            print(f"[WARN] If shipment has been delivered or has signature, then delivery date is required")
            return False, {}, "If shipment has been delivered or has signature, then delivery date is required"

        references = dynamoson["tables"]["General Info"][0]["Document References"]
        for r in references:
            if r['id'] is None:
                continue
            if not bool(re.compile(r"\S+").search(r['id'])):
                continue
            if r.get('idtype') is None:
                raise Exception(f"{r['id']} had an idtype of {r['idtype']} submitted. A valid idtype must be selected.")

            ref = get_bol_reference_schema(idtype=r["idtype"])
            id_ = r["id"]
            ref.update({
                "id": id_
            })
            main["references"].append(ref)

        entities = dynamoson["tables"]["General Info"][0]["Document Entities"]
        carrier = False
        for e in entities:
            if e.get('type') is None:
                raise Exception(f"Entity {e['name']} had a type of {e['type']} submitted. A valid type must be selected.")
            if e["type"] == "CARRIER":
                carrier = True
            ent = get_bol_entity_schema(type_=e["type"])
            name = e["name"]
            if name is not None:
                name = name.upper()
            city = e["city"]
            if city is not None:
                city = city.upper()
            state = e["state"]
            if state is not None:
                state_success, state, state_err = format_state(state=state, abbreviate=True)
                if not state_success:
                    raise Exception(f"format_state() >>> {state_err}")

            country = e.get("country")
            if country is not None:
                country = country.upper()
            address = e["address"]
            if address is not None:
                address = address.upper()
            postal = e["postal"]
            if postal is not None:
                postal = str(postal)

            contactname = e["contactname"]
            if contactname is not None:
                contactname = contactname.upper()
            contact_type = e.get("contact_type")
            if contact_type is not None:
                contact_type = contact_type.upper()
            contact_number = e["contact_number"]
            if contact_number is not None:
                contact_number = contact_number.upper()
            contact_number_type = e.get("contact_number_type")
            if contact_number_type is not None:
                contact_number_type = contact_number_type.upper()

            ent.update({
                "name": name,
                "city": city,
                "state": state,
                "country": country,
                "postal": postal,
                "address": [address],
                "contacts": {
                    "contactname": contactname,
                    "contact_type": contact_type,
                    "contact_number": contact_number,
                    "contact_number_type": contact_number_type
                }
            })
            main["entities"].append(ent)

        if carrier is False:
            print(f"[FAIL] Carrier not found!")
            return False, main, "Carrier is a required entity"

        return True, main, ""

    except Exception as e:
        print(f"[ERROR] {func} ==> {e}")
        traceback.print_exc()
        return False, {}, f"{e}"



import json
import traceback
from dukeai_lib.utilities import DecimalEncoder


def get_main_schema():
    schema = {
        "headers": {
            "Header": {
                "file_type": "",
                "uploaded": "",
                "company_name": ""
            }
        },
        "tables": {
            "Item": [
                {
                    "file_type": "",
                    "transactionDate": "",
                    "category": "",
                    "MainCategory": "",
                    "assetType": "",
                    "amount": "",
                    "state": "",
                    "gal": "",
                    "description": ""
                }
            ],
            "Totals": [
                {
                    "net": "",
                    "loss": "",
                    "revenue": ""
                }
            ]
        }
    }
    return schema


def flatten_data_rsi(ddb_obj: dict) -> tuple[bool, dict, str]:
    """
    Takes a DDB object and flattens data into Dynamotator schema;
    :param ddb_obj: dict, DUKE-User-Document object;
    :return: str, flattened schema for Dynamotator UI as json string.
    """
    func = flatten_data_rsi.__name__
    try:
        ddb_copy = ddb_obj.copy()
        auto_ex = ddb_obj.get("autoextracts")
        main_dict = get_main_schema()

        file_type = ddb_obj.get("file_type")
        uploaded = ddb_obj.get("uploaded")
        c_name = ddb_obj.get("company_name")

        main_dict["headers"]["Header"].update({
            "file_type": file_type,
            "uploaded": uploaded,
            "company_name": c_name
        })

        if c_name is None:
            print(f"[WARN] DDB entry has no company_name, unable to locate info in data array; Defaulting to empty schema")
            main_dict["headers"]["Header"].update({"company_name": ""})
            return True, main_dict, "DDB entry has no company_name, unable to locate info in data array; Defaulting to empty schema"

        else:

            if auto_ex is None:
                print(f"[WARN] DDB entry has no autoextracts; Defaulting to empty schema")
                return True, main_dict, "DDB entry has no autoextracts; Defaulting to empty schema"
            else:
                data = auto_ex.get("data")
                if data is None:
                    auto_ex.update({'data': []})
                    data = auto_ex.get("data")

                else:
                    settlement = False
                    key = c_name + "_none"
                    info = data[0].get(key)
                    dat1 = info.get("Invoice")
                    dat2 = info.get("Receipt")
                    if len(dat1) > 0 and len(dat2) > 0:
                        main_dict["headers"]["Header"].update({"file_type": "settlement"})
                        settlement = True

                    if settlement is False:
                        if len(dat1) > 0:  # Invoice
                            dat1 = dat1[0]
                            amount = dat1.get("amount")
                            if amount is not None:
                                amount = float(amount)
                            gal = dat1.get("gal")
                            desc = dat1.get("description")
                            state = dat1.get("state")
                            cat = dat1.get("category")
                            m_cat = dat1.get("MainCategory")
                            date = dat1.get("transactionDate")
                            a_type = dat1.get("assetType")

                            main_dict["tables"]["Item"][0].update({
                                "file_type": file_type,
                                "transactionDate": date,
                                "category": cat,
                                "MainCategory": m_cat,
                                "assetType": a_type,
                                "amount": amount,
                                "state": state,
                                "gal": gal,
                                "description": desc
                            })
                            for k, v in main_dict["tables"]["Item"][0].items():
                                if v is None:
                                    main_dict["tables"]["Item"][0].update({k: ""})
                                else:
                                    continue

                            main_dict["tables"]["Totals"][0].update({
                                "net": amount,
                                "loss": "",
                                "revenue": amount
                            })
                            for k, v in main_dict["tables"]["Totals"][0].items():
                                if v is None:
                                    main_dict["tables"]["Totals"][0].update({k: ""})
                                else:
                                    continue

                            try:
                                flattened_data = json.loads(json.dumps(main_dict, indent=2, cls=DecimalEncoder, separators=(',', ': ')))
                                # flattened_data = json.loads(json.dumps(main_dict, indent=2, parse_floats=Decimal, separators=(',', ': ')))
                                ddb_copy["autoextracts"] = flattened_data
                                return True, ddb_copy, ""
                            except Exception as e:
                                print(f"[FAIL] Dictionary contains characters that are illegal JSON syntax; Defaulting to blank schema; ERROR = {e}")
                                traceback.print_exc()
                                fail = get_main_schema()
                                return False, fail, f"{e}"

                        elif len(dat2) > 0:  # Receipt
                            dat1 = dat2[0]
                            amount = dat1.get("amount")
                            if amount is not None:
                                amount = float(amount)
                            gal = dat1.get("gal")
                            desc = dat1.get("description")
                            state = dat1.get("state")
                            cat = dat1.get("category")
                            m_cat = dat1.get("MainCategory")
                            date = dat1.get("transactionDate")
                            a_type = dat1.get("assetType")

                            main_dict["tables"]["Item"][0].update({
                                "file_type": file_type,
                                "category": cat,
                                "MainCategory": m_cat,
                                "assetType": a_type,
                                "amount": amount,
                                "state": state,
                                "gal": gal,
                                "description": desc
                            })
                            for k, v in main_dict["tables"]["Item"][0].items():
                                if v is None:
                                    main_dict["tables"]["Item"][0].update({k: ""})
                                else:
                                    continue

                            main_dict["tables"]["Totals"][0].update({
                                "net": amount,
                                "loss": "",
                                "revenue": amount
                            })
                            for k, v in main_dict["tables"]["Totals"][0].items():
                                if v is None:
                                    main_dict["tables"]["Totals"][0].update({k: ""})
                                else:
                                    continue

                            try:
                                flattened_data = json.loads(json.dumps(main_dict, indent=2, cls=DecimalEncoder, separators=(',', ': ')))
                                # flattened_data = json.loads(json.dumps(main_dict, indent=2, parse_floats=Decimal, separators=(',', ': ')))
                                ddb_copy["autoextracts"] = flattened_data
                                return True, ddb_copy, ""
                            except Exception as e:
                                print(f"[FAIL] Dictionary contains characters that are illegal JSON syntax; Defaulting to blank schema; ERROR = {e}")
                                traceback.print_exc()
                                fail = get_main_schema()
                                return False, fail, f"{e}"

                    elif settlement is True:
                        print(f"[INFO] settlement is True; passing for now")
                        return True, main_dict, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        try:
            fail_dict = get_main_schema()
            return False, fail_dict, f"{e}"
        except Exception as ee:
            print(f"[DOUBLE FAIL] {func} failed to generate blank schema; Error ==> {ee}; Original Error ==> {e}")
            traceback.print_exc()
            return False, {}, f"{ee}"



import json
import decimal


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            if o % 1 > 0:
                return float(o)
            else:
                return float(o)
        return super(DecimalEncoder, self).default(o)



from setuptools import setup, find_packages

setup(
    name='dukeai_lib',
    version="0.2.2",
    description="Common functions used across the DUKE.ai project environments.",
    url='https://duke.ai',
    author='Blake Donahoo',
    author_email='blake@duke.ai',
    license='GNU General Public License v3 (GPLv3)',
    packages=find_packages(),
    keywords=['dukeai', 'duke.ai'],
    install_requires=[
        'chalice~=1.27.1',
        'requests~=2.27.1',
        'base58==2.1.1',
        'urllib3==1.26.9'
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Utilities"
    ],
)

# python3 -m build
# python3 -m twine upload --repository duke-pypi dist/dukeai_lib-0.2.0*




import json
import traceback
import re
import datetime
from dukeai_lib.utilities import DecimalEncoder
from .schema_utilities import get_idtype, REFERENCE_OPTIONS, STATES


def get_main_schema():
    main_schema = {
        "headers": {
            "General": {},
            "Broker Details": {},
        },
        "tables": {
            "General Info": [
                {
                    "Document Dates": [],
                    "Shipment Details": [],
                    "Document References": [],
                    "Document Notes": []
                }
            ],
            "Stops": []
        }
    }
    return main_schema


def get_flat_stop_schema():
    # main_schema >> Stops
    stop_schema = {
        "Stop Type": [],
        "Dates": [],
        "Entity": [],
        "References": [],
        "Order Detail": [],
        "Notes": []
    }
    return stop_schema


def get_general_info():
    # main_schema >> headers >> General
    general_info = {
        "identifier": "",
        "identifier_type": "ID Number",
        "client": "",
        "sender": "",
        "receiver": "",
        "isa_ID": ""
    }
    return general_info


def get_entity():
    # main_schema >> headers >> Broker Details
    # stop_schema >> Entity
    entity = {
        "type": "",
        "_type": "",
        "name": "",
        "contactname": "",
        "contact_number": "",
        "contact_type": "",  # Non-rendering
        "contact_number_type": "",  # Non-rendering
        "address": [],
        "city": "",
        "state": "",
        "postal": ""
    }
    return entity


def get_stoptype():
    # stop_schema >> Stop Type
    stop_type = {
        "stoptype": "",
        "_stoptype": "",
        "ordinal": ""  # non-rendering // integer
    }
    return stop_type


def get_date():
    # main_schema >> tables >> General Info >> Document Dates
    # stop_schema >> Dates
    dates = {
        "date_0": "",  # yyyy-mm-dd earliest
        "time_0": "",  # 00:00 earliest
        "date_1": "",  # yyyy-mm-dd latest
        "time_1": "",  # 00:00 latest
        "datetype_0": "",  # Non-rendering
        "timetype_0": "",  # Non-rendering
        "datetype_1": "",  # Non-rendering
        "timetype_1": ""  # Non-rendering
    }
    return dates


def get_shipment():
    # main_schema >> tables >> General Info >> Shipment Details
    shipment = {
        "equipment_number": "",
        "distance": "",
        "weight": "",
        "weight_unit_code": "LBS.",
        "weight_qualifier": "",
        "volume": "",
        "volume_qualifier": "",
        "truck_type": "",
        "temperature": "",
        "trucklength": "",
        "charges": "",
        "loading_quantity": ""
    }
    return shipment


def get_reference():
    # main_schema >> tables >> General Info >> Document References
    # stop_schema >> References
    reference = {
        "id": "",
        "_idtype": "",
        "idtype": ""  # Non-rendering
    }
    return reference


def get_note():
    # main_schema >> tables >> General Info >> Document Notes
    # stop_schema >> Notes
    note = {
        "note": "",
        "notetype": "",  # Non-rendering
        "_notetype": ""  # Non-rendering
    }
    return note


def get_purchase_order():
    # stop_schema >> Order Detail
    purchase_order = {
        "purchase_order_number": "",
        "date": "",
        "cases": "",
        "weight": "",
        "weight_unit_code": "",
        "volume_type": "",
        "volume_units": ""
    }
    return purchase_order


def create_blank_stop(stop_type: str, ordinal: int):
    try:
        stop_dict = get_flat_stop_schema()
        stop_type_dict = dict(get_stoptype())
        if stop_type.lower() == "pick":
            stop_type_dict.update({"stoptype": "PICK"})
            stop_type_dict.update({"ordinal": ordinal})
        elif stop_type.lower() == "drop":
            stop_type_dict.update({"stoptype": "DROP"})
            stop_type_dict.update({"ordinal": ordinal})
        stop_dict["Stop Type"].append(stop_type_dict)
        stop_dict["Dates"].append(get_date())
        entity = get_entity()
        if stop_type.lower() == "pick":
            entity.update({'type': 'SHIPPER'})
        if stop_type.lower() == "drop":
            entity.update({'type': 'CONSIGNEE'})
        stop_dict["Entity"].append(entity)
        stop_dict["References"].append(get_reference())
        stop_dict["Order Detail"].append(get_purchase_order())
        stop_dict["Notes"].append(get_note())
        return stop_dict
    except Exception as e:
        print(e)
        stop_dict = get_flat_stop_schema()
        stop_type_dict = get_stoptype()
        stop_dict["Stop Type"].append(stop_type_dict)
        stop_dict["Dates"].append(get_date())
        stop_dict["Entity"].append(get_entity())
        stop_dict["References"].append(get_reference())
        stop_dict["Order Detail"].append(get_purchase_order())
        stop_dict["Notes"].append(get_note())
        return stop_dict


def fail_schema():
    main_dict = get_main_schema()
    broker = get_entity()
    broker.update({'type': 'BROKER'})
    main_dict["headers"].update({"General": dict(get_general_info()), "Broker Details": dict(broker)})
    general = main_dict["tables"]["General Info"][0]
    general["Document Dates"].append(get_date())
    general["Shipment Details"].append(get_shipment())
    general["Document References"].append(get_reference())
    general["Document Notes"].append(get_note())
    stops = main_dict["tables"]["Stops"]
    pick = create_blank_stop("PICK", 1)
    stops.append(pick)
    drop = create_blank_stop("DROP", 2)
    stops.append(drop)

    return main_dict


def flatten_rate_confirmation(ddb_obj: dict) -> tuple[bool, dict, str]:
    """
    Takes a DDB object and flattens data into Dynamotator schema.
    :param ddb_obj: dict, DUKE-User-Document object.
    :return: str, flattened schema for Dynamotator UI as json string.
    """
    # known special characters to remove that don't play nicely with JSON
    func = flatten_rate_confirmation.__name__
    char_dict = r"\"|\'|\x0c"
    ref_options = list(REFERENCE_OPTIONS.keys())
    try:
        # ddb_obj = json.loads(ddb_str)
        ddb_copy = ddb_obj.copy()
        auto_ex = ddb_obj.get("autoextracts")

        if auto_ex is None:
            print(f"[WARN] DDB entry has no autoextracts; Defaulting to empty schema")
            main_dict = fail_schema()
            return True, main_dict, ""

        else:
            data = auto_ex.get("data")
            if data is None:
                main_dict = get_main_schema()
                general_info = get_general_info()
                auto_ex.update({'data': {}})
                data = auto_ex.get("data")
            else:
                main_dict = get_main_schema()
                # main_dict >> headers >> General
                general_info = get_general_info()
                sender = data.get("sender")
                client = data.get("client")
                receiver = data.get("receiver")
                if receiver is None:
                    data.update({
                        "receiver": {
                            "name": "",
                            "isa_qual": "ZZ",
                            "isa_ID": ""
                        }
                    })
                    receiver = data.get("receiver")
                rec_name = receiver.get("name")

                isa_id = receiver.get("isa_ID")

                identifier = data.get("identifier")
                identifier_type = data.get("identifier_type")
                general_info.update({
                    "identifier": identifier,
                    # "identifier_type": identifier_type,
                    "client": client,
                    "sender": sender,
                    "receiver": rec_name,
                    "isa_ID": isa_id
                })

                for k, v in general_info.items():
                    if v is None:
                        general_info.update({k: ""})

            main_dict["headers"].update({"General": dict(general_info)})
            # main_dict >> headers >> Broker Details
            broker = False
            broker_index = 0
            broker_info = get_entity()
            entities = data.get("entities")
            if entities is None:
                print(f"[INFO] No entities found for this document.")
                broker = False

            else:
                for ent in entities:
                    type_ = ent.get("type")
                    if type_.lower() != "broker":
                        name = ent.get("name")
                        print(f"[INFO] Passing over entity ({name}) of type ({type_}), only looking for broker")

                    else:
                        broker_index = entities.index(ent)
                        broker = True

            if broker is False:
                for k, v in broker_info.items():
                    if isinstance(v, list):
                        if len(v) == 0:
                            broker_info.update({k: [""]})

                    elif v is None:
                        broker_info.update({k: ""})

                main_dict["headers"].update({"Broker Details": dict(broker_info)})
                print("[WARN] could not find broker")

            else:
                entity = entities[broker_index]
                contactname = ""
                contact_number = ""
                contact_type = ""
                contact_number_type = ""
                name = entity.get("name")
                city = entity.get("city")
                state = entity.get("state")
                postal = entity.get("postal")
                address = entity.get("address")
                contacts = entity.get("contacts")
                if contacts is not None:
                    contactname = contacts.get("contactname")
                    contact_number = contacts.get("contact_number")
                    contact_type = contacts.get("contact_type")
                    contact_number_type = contacts.get("contact_number_type")

                broker_info.update({
                    "name": name,
                    "contactname": contactname,
                    "contact_number": contact_number,
                    "contact_number_type": contact_number_type,  # non-rendering
                    "contact_type": contact_type,  # non-rendering
                    "city": city,
                    "state": state,
                    "postal": postal,
                    "type": "BROKER"
                })
                if isinstance(broker_info["address"], list):
                    if len(address) > 1:
                        broker_info["address"] = []
                        for i in address:
                            if isinstance(i, list):
                                i = "".join(i)
                            i = re.sub(char_dict, "", i)
                            broker_info["address"].append(i)
                    elif len(address) == 1:
                        broker_info["address"] = []
                        if isinstance(address[0], list):
                            address[0] = "".join(address[0])
                        ii = re.sub(char_dict, "", address[0])
                        broker_info["address"].append(ii)
                else:
                    print(f"[WARN] Extracted type of address is not list; address = {address}")

                for k, v in broker_info.items():
                    if isinstance(v, list):
                        if len(v) == 0:
                            broker_info.update({k: []})

                    elif v is None:
                        broker_info.update({k: ""})

            main_dict["headers"].update({"Broker Details": dict(broker_info)})
            general = main_dict["tables"]["General Info"][0]

            # main_dict >> tables >> General Info >> Document Dates
            doc_dates = get_date()
            main_dates = data.get("dates")
            if main_dates is None:
                for k, v in doc_dates.items():
                    if v is None:
                        doc_dates.update({k: ""})

                general["Document Dates"].append(doc_dates)
            elif len(main_dates) == 1:
                date = main_dates[0].get("date")
                time = main_dates[0].get("time")
                datetype_0 = main_dates[0].get("datetype")  # non-rendering
                timetype_0 = main_dates[0].get("timetype")  # non-rendering
                doc_dates.update({
                    "date_0": date,
                    "time_0": time,
                    "datetype_0": datetype_0,  # non-rendering
                    "timetype_0": timetype_0  # non-rendering
                })
                for k, v in doc_dates.items():
                    if v is None:
                        doc_dates.update({k: ""})

                general["Document Dates"].append(doc_dates)
            elif len(main_dates) == 2:
                date1 = main_dates[0].get("date")
                time1 = main_dates[0].get("time")
                datetype_0 = main_dates[0].get("datetype")  # non-rendering
                timetype_0 = main_dates[0].get("timetype")  # non-rendering
                date2 = main_dates[1].get("date")
                time2 = main_dates[1].get("time")
                datetype_1 = main_dates[1].get("datetype")  # non-rendering
                timetype_1 = main_dates[1].get("timetype")  # non-rendering
                doc_dates.update({
                    "date_0": date1,
                    "time_0": time1,
                    "datetype_0": datetype_0,  # non-rendering
                    "timetype_0": timetype_0,  # non-rendering
                    "date_1": date2,
                    "time_1": time2,
                    "datetype_1": datetype_1,  # non-rendering
                    "timetype_1": timetype_1  # non-rendering
                })
                for k, v in doc_dates.items():
                    if v is None:
                        doc_dates.update({k: ""})

                general["Document Dates"].append(doc_dates)

            if len(general["Document Dates"]) < 1:
                doc_dates_ = get_date()
                general["Document Dates"].append(doc_dates_)

            # main_dict >> tables >> General Info >> Shipment Details
            shipment = get_shipment()
            main_shipment_details = data.get("shipment")
            if main_shipment_details is None:
                print(f"[INFO] No shipment data key found...is this document completely unclassified? Appending blank shipment schema.")
                general["Shipment Details"].append(shipment)

            else:
                distance = main_shipment_details.get("distance")
                equipment_number = main_shipment_details.get("equipment_number")
                weight = main_shipment_details.get("weight")
                weight_unit_code = main_shipment_details.get("weight_unit_code")
                weight_qualifier = main_shipment_details.get("weight_qualifier")
                volume = main_shipment_details.get("volume")
                volume_qualifier = main_shipment_details.get("volume_qualifier")
                truck_type = main_shipment_details.get("truck_type")
                temperature = main_shipment_details.get("temperature")
                trucklength = main_shipment_details.get("trucklength")
                charges = main_shipment_details.get("charges")
                loading_quantity = main_shipment_details.get("loading_quantity")
                shipment.update({
                    "equipment_number": equipment_number,
                    "distance": distance,
                    "weight": weight,
                    # "weight_unit_code": weight_unit_code,
                    "weight_qualifier": weight_qualifier,
                    "volume": volume,
                    "volume_qualifier": volume_qualifier,
                    "truck_type": truck_type,
                    "temperature": temperature,
                    "trucklength": trucklength,
                    "charges": charges,
                    "loading_quantity": loading_quantity
                })
                for k, v in shipment.items():
                    if v is None:
                        shipment.update({k: ""})

                general["Shipment Details"].append(shipment)

            # main_dict >> tables >> General Info >> Document References
            main_references = data.get("references")
            if main_references is None:
                ref = get_reference()
                general["Document References"].append(ref)

            elif len(main_references) == 0:
                ref = get_reference()
                general["Document References"].append(ref)

            else:
                for reference in main_references:
                    ref = get_reference()
                    id_ = reference.get("id")
                    idtype = reference.get("idtype")
                    if idtype not in ref_options:
                        idtype = "REFERENCE NUMBER"
                    _idtype = get_idtype(idtype)  # non-rendering
                    ref.update({
                        "id": id_,
                        "_idtype": _idtype,
                        "idtype": idtype  # non-rendering
                    })
                    for k, v in ref.items():
                        if v is None:
                            ref.update({k: ""})

                    general["Document References"].append(ref)

            # main_dict >> tables >> General Info >> Document Notes
            main_notes = data.get("notes")
            if main_notes is None:
                note = get_note()
                for k, v in note.items():
                    if v is None:
                        note.update({k: ""})

                general["Document Notes"].append(note)
            elif len(main_notes) == 0:
                note = get_note()
                for k, v in note.items():
                    if v is None:
                        note.update({k: ""})

                general["Document Notes"].append(note)
            else:
                for n in main_notes:
                    note = get_note()
                    # notetype = n.get("notetype")  # non-rendering
                    _notetype = n.get("_notetype")  # non-rendering
                    msg = n.get("note")
                    if msg is not None:
                        msg = re.sub(char_dict, "", msg)
                    else:
                        msg = ""

                    note.update({
                        "note": msg,
                        "notetype": "RATECON NOTES",  # non-rendering
                        "_notetype": "RN"  # non-rendering
                    })
                    for k, v in note.items():
                        if v is None:
                            note.update({k: ""})

                    general["Document Notes"].append(note)

            # main_schema >> Stops
            stops_data = data.get("stops")
            stop_count = 0
            if stops_data is None:
                stop1 = create_blank_stop("PICK", 1)
                stop2 = create_blank_stop("DROP", 2)
                main_dict["tables"]["Stops"].append(stop1)
                main_dict["tables"]["Stops"].append(stop2)
            elif len(stops_data) == 0:
                stop1 = create_blank_stop("PICK", 1)
                stop2 = create_blank_stop("DROP", 2)
                main_dict["tables"]["Stops"].append(stop1)
                main_dict["tables"]["Stops"].append(stop2)
            else:
                count = 1
                for stop in stops_data:
                    stops_dict = get_flat_stop_schema()
                    # Stop Type
                    stoptype_dict = get_stoptype()
                    stoptype = stop.get("stoptype")
                    _stoptype = stop.get("_stoptype")
                    ordinal = stop.get("ordinal")
                    if ordinal is None:
                        ordinal = count
                    if stoptype is not None:
                        stoptype_dict.update({
                            "stoptype": stoptype,
                            "_stoptype": _stoptype,  # non-rendering
                            "ordinal": ordinal
                        })
                    else:
                        print("[WARN] stoptype not found!")
                    stops_dict["Stop Type"].append(stoptype_dict)
                    # Dates
                    date_dict = get_date()
                    stop_dates = stop.get("dates")
                    if stop_dates is None:
                        for k, v in date_dict.items():
                            if v is None:
                                date_dict.update({k: ""})

                        stops_dict["Dates"].append(date_dict)
                        print(f"[WARN] No stop-level dates for stop number {stop_count}")
                    elif len(stop_dates) > 0:
                        if len(stop_dates) == 1:
                            date = stop_dates[0].get("date")
                            time = stop_dates[0].get("time")
                            datetype_0 = stop_dates[0].get("datetype")  # non-rendering
                            timetype_0 = stop_dates[0].get("timetype")  # non-rendering
                            date_dict.update({
                                "date_0": date,
                                "time_0": time,
                                "datetype_0": datetype_0,  # non-rendering
                                "timetype_0": timetype_0  # non-rendering
                            })
                            for k, v in date_dict.items():
                                if v is None:
                                    date_dict.update({k: ""})

                            stops_dict["Dates"].append(date_dict)
                        elif len(stop_dates) == 2:
                            date1 = stop_dates[0].get("date")
                            time1 = stop_dates[0].get("time")
                            datetype_0 = stop_dates[0].get("datetype")  # non-rendering
                            timetype_0 = stop_dates[0].get("timetype")  # non-rendering
                            date2 = stop_dates[1].get("date")
                            time2 = stop_dates[1].get("time")
                            datetype_1 = stop_dates[1].get("datetype")  # non-rendering
                            timetype_1 = stop_dates[1].get("timetype")  # non-rendering
                            date_dict.update({
                                "date_0": date1,
                                "time_0": time1,
                                "datetype_0": datetype_0,  # non-rendering
                                "timetype_0": timetype_0,  # non-rendering
                                "date_1": date2,
                                "time_1": time2,
                                "datetype_1": datetype_1,  # non-rendering
                                "timetype_1": timetype_1  # non-rendering
                            })
                            for k, v in date_dict.items():
                                if v is None:
                                    date_dict.update({k: ""})

                            stops_dict["Dates"].append(date_dict)
                    # Entity
                    entity_dict = get_entity()
                    entity = stop.get("entities")
                    if entity is not None:
                        if len(entity) == 0:
                            entity.append(get_entity())
                        contactname = ""
                        contact_number = ""
                        name = entity[0].get("name")
                        city = entity[0].get("city")
                        state = entity[0].get("state")
                        postal = entity[0].get("postal")
                        address = entity[0].get("address")
                        contacts = entity[0].get("contacts")
                        type_ = entity[0].get("type")
                        _type = entity[0].get("_type")
                        if contacts is not None:
                            contactname = contacts.get("contactname")
                            contact_number = contacts.get("contact_number")
                        entity_dict.update({
                            "type": type_,
                            "_type": _type,
                            "name": name,
                            "contactname": contactname,
                            "contact_number": contact_number,
                            "city": city,
                            "state": state,
                            "postal": postal
                        })
                        if isinstance(address, list):
                            if len(address) > 1:
                                entity_dict["address"] = []
                                for i in address:
                                    if isinstance(i, list):
                                        i = "".join(i)
                                    i = re.sub(char_dict, "", i)
                                    entity_dict["address"].append(i)
                            elif len(address) == 1:
                                entity_dict["address"] = []
                                if isinstance(address[0], list):
                                    address[0] = "".join(address[0])
                                ii = re.sub(char_dict, "", address[0])
                                entity_dict["address"].append(ii)
                        else:
                            print(f"[WARN] Extracted type of address is not list; address = {address}")
                        for k, v in entity_dict.items():
                            if isinstance(v, list):
                                if len(v) == 0:
                                    entity_dict.update({k: []})

                            elif v is None:
                                entity_dict.update({k: ""})

                        stops_dict["Entity"].append(entity_dict)
                    else:
                        for k, v in entity_dict.items():
                            if isinstance(v, list):
                                if len(v) == 0:
                                    entity_dict.update({k: [""]})

                            elif v is None:
                                entity_dict.update({k: ""})

                        stops_dict["Entity"].append(entity_dict)

                    # References
                    stop_references = stop.get("references")
                    if stop_references is None:
                        ref = get_reference()
                        for k, v in ref.items():
                            if v is None:
                                ref.update({k: ""})

                        stops_dict["References"].append(ref)
                    elif len(stop_references) == 0:
                        ref = get_reference()
                        for k, v in ref.items():
                            if v is None:
                                ref.update({k: ""})

                        stops_dict["References"].append(ref)
                    else:
                        for reference in stop_references:
                            ref = get_reference()
                            id_ = reference.get("id")
                            idtype = reference.get("idtype")
                            if idtype not in ref_options:
                                idtype = "REFERENCE NUMBER"
                            _idtype = get_idtype(idtype)  # non-rendering
                            ref.update({
                                "id": id_,
                                "_idtype": _idtype,
                                "idtype": idtype  # non-rendering
                            })
                            for k, v in ref.items():
                                if v is None:
                                    ref.update({k: ""})

                            stops_dict["References"].append(ref)

                    # Order Detail
                    stop_order_detail = stop.get("order_detail")
                    if len(stop_order_detail) == 0:
                        po = get_purchase_order()
                        for k, v in po.items():
                            if v is None:
                                po.update({k: ""})
                            else:
                                continue
                        stops_dict["Order Detail"].append(po)
                    else:
                        if len(stop_order_detail) > 0:
                            for p in stop_order_detail:
                                po = get_purchase_order()
                                po_number = p.get("purchase_order_number")
                                date = p.get("date")
                                cases = p.get("cases")
                                weight_unit_code = p.get("weight_unit_code")
                                weight = p.get("weight")
                                volume_type = p.get("volume_type")
                                volume_units = p.get("volume_units")
                                po.update({
                                    "purchase_order_number": po_number,
                                    "date": date,
                                    "cases": cases,
                                    "weight": weight,
                                    "weight_unit_code": weight_unit_code,
                                    "volume_type": volume_type,
                                    "volume_units": volume_units
                                })
                                for k, v in po.items():
                                    if v is None:
                                        po.update({k: ""})

                                stops_dict["Order Detail"].append(po)

                    # Notes
                    stop_notes = stop.get("notes")
                    if stop_notes is None:
                        note = get_note()
                        for k, v in note.items():
                            if v is None:
                                note.update({k: ""})

                        stops_dict["Notes"].append(note)
                    elif len(stop_notes) == 0:
                        note = get_note()
                        for k, v in note.items():
                            if v is None:
                                note.update({k: ""})

                        stops_dict["Notes"].append(note)
                    else:
                        for n in stop_notes:
                            note = get_note()
                            if n is None:
                                stops_dict["Notes"].append(note)
                            else:
                                msg = n.get("note")
                                notetype = n.get("notetype")  # non-rendering
                                _notetype = n.get("_notetype")  # non-rendering
                                if msg is not None:
                                    msg = re.sub(char_dict, "", msg)
                                else:
                                    msg = ""

                                note.update({
                                    "note": msg,
                                    "notetype": notetype,  # non-rendering
                                    "_notetype": _notetype  # non-rendering
                                })
                                for k, v in note.items():
                                    if v is None:
                                        note.update({k: ""})

                                stops_dict["Notes"].append(note)
                        count += 1
                    main_dict["tables"]["Stops"].append(stops_dict)
            # Now, to make sure that there are no illegal characters present in our new JSON dict
            # to prevent the frontend from crashing.
            try:
                flattened_data = json.loads(
                    json.dumps(main_dict, indent=2, cls=DecimalEncoder, separators=(',', ': ')))
                ddb_copy["autoextracts"]["data"] = flattened_data
                print(f"[FLATTENED DOC] = {json.dumps(flattened_data)}")
                return True, ddb_copy, ""
            except Exception as je:
                print(f"[DEBUG] json.loads(json.dumps()) could not parse with cls; ERROR = {je}")
                try:
                    flattened_data = json.loads(
                        json.dumps(main_dict, indent=2, separators=(',', ': ')))
                    ddb_copy["autoextracts"]["data"] = flattened_data
                    print(f"[FLATTENED DOC] = {json.dumps(flattened_data)}")
                    return True, ddb_copy, f"{je}"
                except Exception as jeee:
                    print(f"[DEBUG] json.loads(json.dumps()) could not parse data; ERROR = {jeee}")
                    print("[FAIL] Dictionary contains characters that are illegal JSON syntax; Defaulting to blank schema")
                    ddb_copy["autoextracts"]["data"] = fail_schema()
                    return True, ddb_copy, f"{jeee}"

    except Exception as e:
        print(f"[ERROR] {func} failed; Error ==> {e}")
        traceback.print_exc()
        try:
            ddb_obj["autoextracts"]["data"] = fail_schema()
            return False, ddb_obj, f"{e}"
        except Exception as ee:
            print(f"[DOUBLE FAIL] failed to generate blank schema to return; ERROR = {ee}")
            traceback.print_exc()
            return False, {}, f"{ee}"


#  BLANK main DICTIONARY:
def get_rate_confirmation_schema():
    rate_confirmation_schema = {
        "transaction_type": "204",
        "sender": None,  # BROKER or SHIPPER if they are not having a broker. Example: "Werner Logistics"
        "receiver": {
            "name": None,  # carrier-name on top right of page
            "isa_qual": "ZZ",  # hard-coded
            "isa_ID": None  # client email
        },
        "client": None,  # hard-coded. Example: "Werner Logistics"
        "submitted_time": None,  # time we received the email
        "identifier": None,
        "identifier_type": None,
        "shipment": {
            "equipment_number": None,
            "distance": None,
            "weight": None,
            "weight_unit_code": None,
            "weight_qualifier": "GROSS WEIGHT",  # hard-coded
            "volume": None,
            "volume_qualifier": None,
            "truck_type": None,
            "temperature": None,
            "trucklength": None,
            "charges": None,
            "loading_quantity": None
        },
        "purpose": "ORIGINAL",
        "references": [],  # append references_schema here for reference numbers ABOVE Shipper/Consignee
        "dates": [],  # append dates_schema here if any, don't include stop dates
        "notes": [],  # append notes_schema here for notes/comments ABOVE Shipper/Consignee
        "entities": [],  # append entities_schema here
        "stops": [],  # append stops_schema here

    }
    return rate_confirmation_schema


# BLANK references DICTIONARY:
# used for rate-con level references as well as stop-level references
def get_reference_schema():
    reference_schema = {
        "id": None,
        "idtype": None,
        "_idtype": None
    }
    return reference_schema


# BLANK dates DICTIONARY:
def get_dates_schema():
    dates_schema = {
        "date": None,  # dd/mm/yyyy hh:mm
        "datetype": None,  # always "RESPOND BY", "EP", or "LP"?None
        "time": None,
        "timetype": None
        # always "MUST RESPOND BY", "EARLIEST REQUESTED (PICKUP|DROP) TIME", "LATEST REQUESTED (PICKUP|DROP) TIME"?
    }
    return dates_schema


# BLANK references DICTIONARY:
# used for rate-con level notes as well as stop-level notes
def get_note_schema():
    note_schema = {
        "note": None,
        "notetype": None,
        "_notetype": None
    }
    return note_schema


# BLANK entites DICTIONARY VARIABLE:
#  can be Broker, Shipper, or Consignee
def get_entity_schema(entity_type):
    entity_schema = {
        "name": None,
        "type": None,
        "_type": None,
        "id": "",
        "idtype": "MUTUALLY DEFINED",  # hard-coded
        "_idtype": "ZZ",  # hard-coded
        "address": [],  # List object ['address part 1', 'address part 2']
        "city": None,
        "state": None,
        "postal": None,
        "country": None,
        "contacts": {
            "contactname": None,
            "contact_type": None,
            "contact_number": None,
            "contact_number_type": None
        }
    }

    if entity_type.upper() == "SHIPPER":
        entity_schema['type'] = "SHIPPER"
        entity_schema['_type'] = "SH"
        return entity_schema
    elif entity_type.upper() == "CONSIGNEE":
        entity_schema['type'] = "CONSIGNEE"
        entity_schema['_type'] = "CN"
        return entity_schema
    elif entity_type.upper() == "BROKER":
        entity_schema['type'] = "BROKER"
        entity_schema['_type'] = "BK"
        return entity_schema
    else:
        print("Select correct entity type (SHIPPER/CONSIGNEE/BROKER)")
    return entity_schema


# BLANK purchase_order DICTIONARY to be appended to "order_detail" in stops dictionary IF:

# -- if only ONE PO, fill out "stops"["order_detail"] and ignore this extra dictionary.
# -- if multiple PO's use this dictionary and append to "order_detail"["purchase_order_number"] in stops dictionary.
def get_purchase_order_schema():
    purchase_order_schema = {
        "purchase_order_number": None,
        "date": None,
        "cases": None,  # quantity
        "weight_unit_code": None,  # "L" for pounds, "K" for Kilo
        "weight": None,
        "volume_type": None,  # "cubic feet", etc
        "volume_units": None
    }
    return purchase_order_schema


# BLANK stops DICTIONARY:

#  _stopType Codes for Picks & Drops:
#  Picks:
#  LD (Load)   <--- Duke to use this one in general
#  PL (Partial Load)
#  CL (Complete Load)
#  RT (Retrieval of Trailer)

#  Drops:
#  UL (Unload)  <-- Duke to use this one in general
#  PU (Partial Unload)
#  CU (Complete Unload)
#  DT (Drop Trailer)

def get_stops_schema(stop_type, ordinal: int):
    stop_schema = {
        "stoptype": None,  # see stoptype codes for pickups and drops
        "_stoptype": None,  # see stoptype codes for pickups and drops
        "ordinal": ordinal,  # starts from 1 for first "PICK"
        "dates": [],  # append dates_schema here
        "references": [],  # append references_schema here for stop references
        "order_detail": [],
        "entities": [],  # append entities_schema here for stop-level entities
        "notes": []  # append notes_schema here for stop-level notes/comments
    }

    if stop_type.upper() == "PICK":
        stop_schema['stoptype'] = "PICK"
        stop_schema['_stoptype'] = "LD"
        return stop_schema
    elif stop_type.upper() == "DROP":
        stop_schema['stoptype'] = "DROP"
        stop_schema['_stoptype'] = "UL"
        return stop_schema
    else:
        print("Select correct entity type (PICK/DROP)")
    return stop_schema


def unflatten_ratecon_time(time_str):
    func = unflatten_ratecon_time.__name__
    try:
        if time_str is None:
            return True, None, ""

        elif not bool(re.compile(r"\S").search(time_str)):
            print(f"Time is empty string, skipping...")
            return True, None, ""

        else:

            try:
                time_ = datetime.datetime.strptime(time_str, '%H:%M')
                time_0 = time_.strftime('%H:%M')
            except Exception as te:
                print(f"[{func}] Warning ==> {te}")

                try:
                    time_ = datetime.datetime.strptime(time_str, '%H:%M:%S')
                    time_0 = time_.strftime('%H:%M')
                except Exception as te2:
                    print(f"[{func}] Warning ==> {te2}")

                    try:
                        time_ = datetime.datetime.strptime(time_str, '%H:%M:%S.%f')
                        time_0 = time_.strftime('%H:%M')
                    except Exception as te3:
                        print(f"[ERROR PARSING TIME] submitted time: {time_str} ==> {te3}")
                        raise Exception(f"Time ({time_str}) is incorrect format >>> {te3}")

            return True, time_0, ""
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"


def unflatten_ratecon_date(date_str):
    func = unflatten_ratecon_date.__name__
    try:
        if date_str is None:
            return True, None, ""

        elif not bool(re.compile(r"\S").search(date_str)):
            print(f"Date is empty string, skipping...")
            return True, None, ""

        elif date_str.lower() in ["n/a", "na"]:
            return True, None, ""

        else:
            try:
                date_ = datetime.datetime.strptime(date_str, '%Y-%m-%d')
                date_0 = date_.strftime('%Y-%m-%d')
            except Exception as de:
                raise Exception(f"Date submitted ({date_str}) is incorrectly formatted. Should be YYYY-mm-dd; Error ==> {de}")

            return True, date_0, ""
    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, None, f"{e}"


def unflatten_rate_confirmation(flattened_ddb_entry: str) -> tuple[bool, dict, str]:
    """
    Takes flattened data from the Dynamotator UI and rebuilds the information into the original Ratecon Schema;
    :param flattened_ddb_entry: JSON str
    :return: JSON str, Original ratecon schema inside the ddb_entry.
    """
    func = unflatten_rate_confirmation.__name__
    if isinstance(flattened_ddb_entry, dict):
        flattened_ddb_entry = json.dumps(flattened_ddb_entry)

    data = json.loads(flattened_ddb_entry)
    main = get_rate_confirmation_schema()
    try:
        general = data["headers"]["General"]
        identifier = general.get("identifier")
        identifier_type = general.get("identifier_type")
        client = general.get("client")
        sender = general.get("sender")
        receiver = general.get("receiver")
        isa_id = general.get("isa_ID")
        main.update({
            "client": client,
            "sender": sender,
            "identifier": identifier,
            "identifier_type": identifier_type
        })
        main["receiver"].update({
            "name": receiver,
            "isa_ID": isa_id
        })

        ship = data["tables"]["General Info"][0]["Shipment Details"][0]
        distance = ship.get("distance")
        equipment_number = ship.get("equipment_number")
        weight = ship.get("weight")
        weight_unit_code = ship.get("weight_unit_code")
        weight_qualifier = ship.get("weight_qualifier")
        volume = ship.get("volume")
        volume_qualifier = ship.get("volume_qualifier")
        truck_type = ship.get("truck_type")
        temperature = ship.get("temperature")
        trucklength = ship.get("trucklength")
        charges = ship.get("charges")
        if charges is not None:
            if charges.lower() in ["na", "n/a"]:
                charges = "0.00"
            else:
                charges = re.sub(r"[^\d.,]+", "", charges)
        loading_quantity = ship.get("loading_quantity")
        main["shipment"].update({
            "distance": distance,
            "equipment_number": equipment_number,
            "weight": str(weight),
            "weight_unit_code": weight_unit_code,
            "weight_qualifier": weight_qualifier,
            "volume": volume,
            "volume_qualifier": volume_qualifier,
            "truck_type": truck_type,
            "temperature": temperature,
            "trucklength": trucklength,
            "charges": charges,
            "loading_quantity": loading_quantity
        })
        for k, v in main['shipment'].items():
            if v is None or not bool(re.compile(r"\S").search(str(v))):
                main['shipment'].update({k: None})

        flt_dates = data["tables"]["General Info"][0]["Document Dates"]
        date_dict_0 = get_dates_schema()
        for i in flt_dates:

            date_success, date_0, date_err = unflatten_ratecon_date(date_str=i.get('date_0'))
            if not date_success:
                raise Exception(f"unflatten_ratecon_date() >>> {date_err}")

            time_success, time_0, time_err = unflatten_ratecon_time(time_str=i.get('time_0'))
            if not time_success:
                raise Exception(f"unflatten_ratecon_time() >>> {time_err}")

            datetype_0 = i.get("datetype_0")
            timetype_0 = i.get("timetype_0")
            date_dict_0.update({
                "date": date_0,
                "datetype": datetype_0,
                "time": time_0,
                "timetype": timetype_0
            })
            for k, v in date_dict_0.items():
                if v is None or not bool(re.compile(r"\S").search(str(v))):
                    date_dict_0.update({k: None})

            if re.compile(r"\d").search(str(date_dict_0["date"])) or re.compile(r"\d").search(str(date_dict_0["time"])):
                main["dates"].append(date_dict_0)
            else:
                print(f"[INFO] Passing over date in Document Dates, no values found")
                main["dates"] = []

        flt_note = data["tables"]["General Info"][0]["Document Notes"]
        if isinstance(flt_note, list):
            if len(flt_note) == 0:
                print("[INFO] No Document-level notes found")
                main["notes"] = []
            else:
                for i in flt_note:
                    note = dict(get_note_schema())
                    msg = i.get("note")
                    note.update({
                        "note": msg,
                        "notetype": "RATECON NOTES",
                        "_notetype": "RN"
                    })
                    if note['note'] is not None:
                        if re.compile(r"[a-zA-Z0-9]").search(note["note"]):
                            main["notes"].append(note)
                        else:
                            print(f"[INFO] No Document-level notes found")
                            continue
                    else:
                        print(f"[INFO] No Document-level notes found")
                        continue
        else:
            print("[WARN] Incoming type of notes from tables>>Document Notes does not match expected datatype of list")
            main["notes"] = []

        broker_dict = get_entity_schema("BROKER")
        flt_broker = data["headers"]["Broker Details"]
        name_ = "None"
        city_ = "None"
        state_ = "None"
        postal_ = "None"
        flt_address = flt_broker.get("address")
        name = flt_broker.get("name")
        if name is not None:
            name = name.upper()
            name_ = name
        city = flt_broker.get("city")
        if city is not None:
            city = city.upper()
            city_ = city
        state = flt_broker.get("state")
        if state is not None:
            if isinstance(state, str):
                for name, abbrev in STATES.items():
                    if state.lower() == name.lower():
                        print(f"[STATE ABBREV CONVERSION] state={state}; converted to: {abbrev.upper()}")
                        state = abbrev.upper()
                    elif state.lower() == abbrev.lower():
                        print(f"[STATE ABBREV MATCHED] state={state.upper()}")
                        state = abbrev.upper()

        postal = flt_broker.get("postal")
        if postal is not None:
            postal_ = postal
        uid_ = name_ + city_ + state + postal_
        uid = re.sub('[^A-Za-z0-9]+', '', uid_)
        contactname = flt_broker.get("contactname")
        contact_type = flt_broker.get("contact_type")
        contact_number = flt_broker.get("contact_number")
        contact_number_type = flt_broker.get("contact_number_type")
        broker_dict.update({
            "id": uid,
            "name": name,
            "city": city,
            "state": state,
            "postal": postal,
            "country": "USA"
        })
        if len(flt_address) == 0:
            print("[INFO] no broker address found!")
            broker_dict["address"] = []
        elif isinstance(flt_address, list):
            if len(flt_address) == 1:
                if isinstance(flt_address[0], list):
                    flt_address[0] = "".join(flt_address[0])
                broker_dict["address"].append(flt_address[0])
            elif len(flt_address) >= 2:
                for i in flt_address:
                    if isinstance(i, list):
                        i = "".join(i)
                    broker_dict["address"].append(i)
        elif isinstance(flt_address, str):
            if len(flt_address) >= 1:
                add_lst = [flt_address]
                broker_dict["address"] = add_lst
            else:
                broker_dict["address"] = []

        broker_dict["contacts"].update({
            "contactname": contactname,
            "contact_type": contact_type,
            "contact_number": contact_number,
            "contact_number_type": contact_number_type
        })
        main["entities"].append(broker_dict)

        flt_references = data["tables"]["General Info"][0]["Document References"]
        if isinstance(flt_references, list):
            if len(flt_references) == 0:
                print("[INFO] No Document-level references found")
                main["references"] = []
            else:
                for i in flt_references:
                    if i.get('id') is None:
                        print(f"skipping NonType reference...")
                        continue
                    if not bool(re.compile(r"\S").search(i.get('id'))):
                        print(f"skipping empty reference...")
                        continue
                    ref = get_reference_schema()
                    _id = i.get("id")
                    idtype = i.get("idtype")
                    _idtype = get_idtype(idtype)
                    ref.update({
                        "id": _id,
                        "_idtype": _idtype,
                        "idtype": idtype
                    })

                    if re.compile(r"[a-zA-Z0-9]").search(str(ref['id'])):
                        main["references"].append(ref)
                    else:
                        main["references"] = []
        else:
            print("[WARN] Incoming type of references from tables>>Document References does not match expected datatype of list")
            main["references"] = []

        flt_stops = data["tables"]["Stops"]
        ordinal = 1
        for s in flt_stops:
            _type = s.get("Stop Type")
            stoptype = _type[0].get("stoptype")
            if stoptype is None or (isinstance(stoptype, str) and stoptype.lower() in ["na", "n/a"]):
                raise Exception(f"Stoptype in stop {ordinal} is blank!")

            stop_dict = get_stops_schema(stoptype, ordinal)

            dates = s["Dates"]
            if len(dates) == 0:
                print(f"[INFO] Passing over primary date in stop {str(ordinal)}, no values found")
                stop_dict["dates"] = []

            elif len(dates) == 1:
                date_dict_0 = dict(get_dates_schema())
                dates = dates[0]

                s_date_success, date_0, s_date_err = unflatten_ratecon_date(date_str=dates.get("date_0"))
                if not s_date_success:
                    raise Exception(f"[Stop {ordinal} - Date 1] unflatten_ratecon_date() >>> {s_date_err}")

                s_time_success, time_0, s_time_err = unflatten_ratecon_time(time_str=dates.get("time_0"))
                if not s_date_success:
                    raise Exception(f"[Stop {ordinal} - Time 1] unflatten_ratecon_time() >>> {s_time_err}")

                datetype_0 = dates.get("datetype_0")
                timetype_0 = dates.get("timetype_0")
                date_dict_0.update({
                    "date": date_0,
                    "datetype": datetype_0,
                    "time": time_0,
                    "timetype": timetype_0
                })
                for k, v in date_dict_0.items():
                    if v is None or not bool(re.compile(r"\S").search(str(v))):
                        date_dict_0.update({k: None})

                if re.compile(r"\d").search(str(date_dict_0["date"])) or re.compile(r"\d").search(str(date_dict_0["time"])):
                    stop_dict["dates"].append(date_dict_0)
                else:
                    stop_dict["dates"] = []

            elif len(dates) >= 2:
                for d in dates:
                    date_dict_1 = dict(get_dates_schema())
                    date_idx = dates.index(d)

                    s_date_success, date_1, s_date_err = unflatten_ratecon_date(date_str=d.get("date_0"))
                    if not s_date_success:
                        raise Exception(f"[Stop {ordinal} - Date {date_idx + 1}] unflatten_ratecon_date() >>> {s_date_err}")

                    s_time_success, time_1, s_time_err = unflatten_ratecon_time(time_str=d.get("time_0"))
                    if not s_date_success:
                        raise Exception(f"[Stop {ordinal} - Time {date_idx + 1}] unflatten_ratecon_time() >>> {s_time_err}")

                    datetype_1 = d.get("datetype_0")
                    timetype_1 = d.get("timetype_0")
                    date_dict_1.update({
                        "date": date_1,
                        "datetype": datetype_1,
                        "time": time_1,
                        "timetype": timetype_1
                    })
                    for k, v in date_dict_1.items():
                        if v is None or not bool(re.compile(r"\S").search(str(v))):
                            date_dict_1.update({k: None})

                    if re.compile(r"\d").search(str(date_dict_1["date"])) or re.compile(r"\d").search(str(date_dict_1["time"])):
                        stop_dict["dates"].append(date_dict_1)
                    else:
                        print(f"[INFO] Passing over date in stop {str(ordinal)}, no values found")
                        continue

            references = s.get("References")
            if isinstance(references, list):
                if len(references) == 0:
                    print(f"[INFO] No Stop-level references found in stop numer {str(ordinal)}")
                    stop_dict["references"] = []
                else:
                    for r in references:
                        if r.get('id') is None:
                            print(f"skipping NonType reference...")
                            continue
                        if not bool(re.compile(r"\S").search(r.get('id'))):
                            print(f"skipping empty reference...")
                            continue
                        ref = dict(get_reference_schema())
                        _id = r.get("id")
                        idtype = r.get("idtype")
                        _idtype = get_idtype(idtype)
                        ref.update({
                            "id": _id,
                            "_idtype": _idtype,
                            "idtype": idtype
                        })

                        if re.compile(r"[a-zA-Z0-9]").search(str(ref['id'])):
                            stop_dict["references"].append(ref)
                        else:
                            continue
            else:
                print(
                    f"[WARN] Incoming type of references from stop number {str(ordinal)} does not match expected datatype of list")
                stop_dict["references"] = []

            flt_notes = s.get("Notes")
            if isinstance(flt_notes, list):
                if len(flt_notes) == 0:
                    print(f"[INFO] No Stop-level notes found in stop number {str(ordinal)}")
                    stop_dict["notes"] = []
                else:
                    for n in flt_notes:
                        note = dict(get_note_schema())
                        msg = n.get("note")
                        note.update({
                            "note": msg,
                            "notetype": "STOP NOTES",
                            "_notetype": "SN"
                        })

                        if note['note'] is not None:
                            if re.compile(r"[a-zA-Z0-9]").search(str(note['note'])):
                                stop_dict["notes"].append(note)
                            else:
                                print(f"[INFO] No note found in stop number {str(ordinal)}")
                                continue
                        else:
                            print(f"[INFO] No note found in stop number {str(ordinal)}")
                            continue
            else:
                print(
                    f"[WARN] Incoming type of notes from stop number {str(ordinal)} does not match expected datatype of list")

            flt_po = s["Order Detail"]
            if isinstance(flt_po, list):
                if len(flt_po) == 0:
                    print(f"[INFO] No Order Detail(s) found in stop number {str(ordinal)}")
                    stop_dict["order_detail"] = [get_purchase_order_schema()]
                else:
                    for p in flt_po:
                        po = get_purchase_order_schema()
                        po_num = p.get("purchase_order_number")
                        date = p.get("date")
                        cases = p.get("cases")
                        weight = p.get("weight")
                        weight_unit_code = p.get("weight_unit_code")
                        volume_type = p.get("volume_type")
                        volume_units = p.get("volume_units")
                        po.update({
                            "purchase_order_number": po_num,
                            "date": date,
                            "cases": cases,
                            "weight": weight,
                            "weight_unit_code": weight_unit_code,
                            "volume_type": volume_type,
                            "volume_units": volume_units
                        })
                        stop_dict["order_detail"].append(po)

            else:
                print(f"[WARN] Incoming type of Order Detail from stop number {str(ordinal)} does not match expected datatype of list")
                stop_dict["order_detail"] = []

            ent = s["Entity"][0]
            _type = ent.get("type")
            if _type is None:
                _type2 = s.get("Stop Type")
                stoptype = _type2[0].get("stoptype")
                if stoptype.lower() == "pick":
                    _type = "SHIPPER"
                elif stoptype.lower() == "drop":
                    _type = "CONSIGNEE"
                elif _type is None:
                    raise Exception("stoptype is not one of [pick, drop]")

            entity = get_entity_schema(_type)
            name_ = "None"
            city_ = "None"
            state_ = "None"
            postal_ = "None"
            flt_address = []
            if ent.get("address") is not None:
                flt_address.append(ent.get("address"))
            name = ent.get("name")
            if name is not None:
                name = name.upper()
                name_ = name
            city = ent.get("city")
            if city is not None:
                city = city.upper()
                city_ = city

            state = ent.get("state")
            if state is not None:
                if isinstance(state, str):
                    for st, abbrev in STATES.items():
                        if state.lower() == st.lower():
                            print(f"[STATE ABBREV CONVERSION] state={state}; converted to: {abbrev.upper()}")
                            state = abbrev.upper()
                        elif state.lower() == abbrev.lower():
                            print(f"[STATE ABBREV MATCHED] state={state.upper()}")
                            state = abbrev.upper()

            postal = ent.get("postal")
            if postal is not None:
                postal_ = postal


            uid_ = name_ + city_ + state + postal_
            uid = re.sub('[^A-Za-z0-9]+', '', uid_)
            contactname = ent.get("contactname")
            contact_number = ent.get("contact_number")
            contact_type = ent.get("contact_type")
            contact_number_type = ent.get("contact_number_type")
            entity.update({
                "id": uid,
                "name": name,
                "city": city,
                "state": state,
                "postal": postal,
                "country": "USA"
            })

            if isinstance(flt_address, list):
                if len(flt_address) == 0:
                    print("[INFO] no entity address found!")
                    entity["address"] = []
                elif len(flt_address) >= 1:
                    for i in flt_address:
                        if isinstance(i, list):
                            i = "".join(i)
                        entity["address"].append(i)
            elif isinstance(flt_address, str):
                if len(flt_address) >= 1:
                    add_lst = [flt_address]
                    entity["address"] = add_lst
                else:
                    entity['address'] = 0

            entity["contacts"].update({
                "contactname": contactname,
                "contact_type": contact_type,
                "contact_number": contact_number,
                "contact_number_type": contact_number_type
            })

            stop_dict["entities"].append(entity)
            main["entities"].append(entity)
            main["stops"].append(stop_dict)
            ordinal += 1

        print(f"TRANSLATED DOCUMENT = {json.dumps(main)}")
        return True, main, ""

    except Exception as e:
        print(f"[ERROR] {func} Error ==> {e}")
        traceback.print_exc()
        return False, {}, f"{e}"

